# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import webbrowser
from threading import Thread
import select
import socket
import logging 
import argparse
from waitress import serve

import odatix.lib.printc as printc
import odatix.lib.term_mode as term_mode
from odatix.explorer.explorer_app import ResultExplorer

######################################
# Settings
######################################

script_name = os.path.basename(__file__)

start_port = 8052
max_find_port_attempts = 50

######################################
# Parse Arguments
######################################

def add_arguments(parser):
  parser.add_argument('-i', '--input', type=str, default='results', help='Directory of the result YAML files')
  parser.add_argument('-n', '--network', action='store_true', help='Run the server on the network')
  parser.add_argument('--normal_term_mode', action='store_true', help='Do not change terminal mode')
  parser.add_argument('--safe_mode', action='store_true', help='Do not exit on internal error')

def parse_arguments():
  parser = argparse.ArgumentParser(description='Odatix - Start Result Explorer')
  add_arguments(parser)
  return parser.parse_args()

######################################
# Misc functions
######################################

def open_browser():
  webbrowser.open("http://" + ip_address + ':' + str(port), new=0, autoraise=True)

def close_server(old_settings):
  if old_settings is not None:
    term_mode.restore_mode(old_settings)
  os._exit(0)

def find_free_port(host, start_port):
  """Find a free port by incrementing from the start_port."""
  sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  port = start_port
  attempts = 0
  while True:
    if attempts >= max_find_port_attempts:
      printc.error(f"Could not find any available port in range [{start_port}:{port}]", script_name)
      sys.exit(-1)
    try:
      sock.bind((host, port))
      sock.close()
      return port
    except OSError:
      port += 1
      attempts += 1

def start_result_explorer(input, network=False, normal_term_mode=False, safe_mode=False):

  global ip_address
  global port

  # Default ip address: local
  host_address = '127.0.0.1'
  ip_address = host_address

  logging.getLogger('waitress').setLevel(logging.ERROR)

  if network:
    host_address = '0.0.0.0'
    ip_address = socket.gethostbyname(socket.gethostname())

  port = find_free_port(host_address, start_port)

  printc.say("Server running on " + printc.colors.BLUE + "http://" + ip_address + ":" + str(port) + '/' + printc.colors.ENDC, end="", script_name=script_name)
  if network:
    print(" (network-accessible)")
  else:
    print(" (localhost only)")

  if normal_term_mode:
    printc.say("press 'q', then enter to quit", script_name=script_name)
  else:
    printc.say("press 'q' to quit", script_name=script_name)

  # Open the web page
  process = Thread(target=open_browser).start()

  if normal_term_mode:
    old_settings = None
  else:
    old_settings = term_mode.set_raw_mode()

  result_explorer = ResultExplorer(
    result_path=input,
    old_settings=old_settings,
    safe_mode=safe_mode
  )

  # Start the server
  serve_thread = Thread(target=serve, args=(result_explorer.app.server,), kwargs={'host': host_address, 'port': port})
  serve_thread.start()
  
  try:
    while True:
      # Check if a key is pressed
      if sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
        key = sys.stdin.read(1).lower()
        if key == 'q':
          close_server(old_settings)
  finally:
    if old_settings is not None:
      term_mode.restore_mode(old_settings)

######################################
# Main
######################################

def main(args=None):
  if args is None:
    args = parse_arguments()

  network = args.network
  input = args.input
  normal_term_mode = args.normal_term_mode
  safe_mode = args.safe_mode
  start_result_explorer(input, network, normal_term_mode, safe_mode)

if __name__ == "__main__":
  main()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State

side_bar_width = "400px"


def top_bar(page_name=""):
  return html.Div(
    [
      html.Div(
        id=f"navbar-title-{page_name}",
        children=[
          dcc.Link("Odatix Explorer", href="/", className="title"),
        ],
        style={"marginLeft": "30px", "transition": "margin-left 0.25s"},
      ),
      html.Div(),
      html.Div(
        [
          dcc.Link("XY", href="/xy", className="nav-link"),
          dcc.Link("VS", href="/vs", className="nav-link"),
          dcc.Link("Radar", href="/radar", className="nav-link"),
          # dcc.Link('Help', href='/help', className='nav-link')
        ],
        className="nav-links",
      ),
    ],
    className="navbar",
  )


def side_bar(content, page_name=""):
  return html.Div(
    [
      html.Div(id=f"sidebar-top-{page_name}", className="sidebar-top", style={"left": "0", "width": side_bar_width}),
      html.Img(
        id=f"toggle-button-{page_name}",
        src="/assets/icons/sidebar-panel-expand-icon.svg",
        n_clicks=0,
        style={
          "display": "none",
          "cursor": "pointer",
          "position": "absolute",
          "top": "10px",
          "left": "20px",
          "width": "30px",
        },
      ),
      html.Div(
        id=f"sidebar-{page_name}",
        children=[
          html.Img(
            id=f"close-button-{page_name}",
            src="/assets/icons/sidebar-panel-collapse-icon.svg",
            n_clicks=0,
            style={"cursor": "pointer", "position": "absolute", "top": "10px", "left": "20px", "width": "30px"},
          ),
          html.Div(children=[content], className="sidebar-content", style={"width": side_bar_width}),
        ],
        className="sidebar",
        style={"left": "0", "width": side_bar_width},
      ),
    ]
  )


def setup_sidebar_callbacks(explorer, page_name=""):
  @explorer.app.callback(
    [
      Output(f"sidebar-{page_name}", "style"),
      Output(f"content-{page_name}", "style"),
      Output(f"navbar-title-{page_name}", "style"),
      Output(f"sidebar-top-{page_name}", "style"),
      Output(f"toggle-button-{page_name}", "style"),
    ],
    [Input(f"toggle-button-{page_name}", "n_clicks"), Input(f"close-button-{page_name}", "n_clicks")],
    [
      State(f"sidebar-{page_name}", "style"),
      State(f"content-{page_name}", "style"),
      State(f"navbar-title-{page_name}", "style"),
      State(f"sidebar-top-{page_name}", "style"),
      State(f"toggle-button-{page_name}", "style"),
    ],
  )
  def toggle_sidebar(
    toggle_n_clicks, close_n_clicks, sidebar_style, content_style, navbar_style, sidebar_top, toggle_style
  ):
    ctx = dash.callback_context
    if not ctx.triggered:
      return sidebar_style, content_style, navbar_style, sidebar_top, toggle_style

    button_id = ctx.triggered[0]["prop_id"].split(".")[0]

    if button_id == f"toggle-button-{page_name}":
      if sidebar_style["left"] == "-" + side_bar_width:
        sidebar_style["left"] = "0"
        sidebar_top["left"] = "0"
        content_style["marginLeft"] = side_bar_width
        navbar_style["marginLeft"] = "30px"
        navbar_style["position"] = "fixed"
        toggle_style["display"] = "none"
      else:
        sidebar_style["left"] = "-" + side_bar_width
        sidebar_top["left"] = "-" + side_bar_width
        content_style["marginLeft"] = "0"
        navbar_style["marginLeft"] = "0"
        navbar_style["position"] = "relative"
        toggle_style["display"] = "block"
    elif button_id == f"close-button-{page_name}":
      sidebar_style["left"] = "-" + side_bar_width
      sidebar_top["left"] = "-" + side_bar_width
      content_style["marginLeft"] = "0"
      navbar_style["marginLeft"] = "0"
      navbar_style["position"] = "relative"
      toggle_style["display"] = "block"

    return sidebar_style, content_style, navbar_style, sidebar_top, toggle_style



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import plotly.graph_objs as go

import odatix.explorer.legend as legend
import odatix.explorer.navigation as navigation

page_name = "vs"


def layout(explorer):
  legend_items = legend.create_legend_items(explorer, page_name)

  return html.Div(
    [
      navigation.top_bar(page_name),
      navigation.side_bar(
        content=html.Div(
          id=f"sidebar-content-{page_name}",
          className="sidebar-content-holder",
          children=[
            html.H2("Data"),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("YAML File")]),
                dcc.Dropdown(
                  id="yaml-dropdown",
                  options=[{"label": yaml_file, "value": yaml_file} for yaml_file in explorer.valid_yaml_files],
                  value=explorer.valid_yaml_files[0] if explorer.valid_yaml_files else None,
                ),
              ],
            ),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Target")]),
                dcc.Dropdown(
                  id=f"target-dropdown-{page_name}",
                  value=explorer.dfs[explorer.valid_yaml_files[0]]["Target"].iloc[0]
                  if explorer.valid_yaml_files
                  else None,
                ),
              ],
            ),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Metric X")]),
                dcc.Dropdown(id="metric-x-dropdown", value="Fmax"),
              ],
            ),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Metric Y")]),
                dcc.Dropdown(id="metric-y-dropdown", value="Fmax"),
              ],
            ),
            html.H2("Architectures"),
            html.Div(
              [
                html.Div(
                  [
                    html.Button("Show All", id="show-all", n_clicks=0),
                    html.Button("Hide All", id="hide-all", n_clicks=0),
                  ]
                ),
                html.Div(legend_items, id="custom-legend", style={"margin-top": "15px", "margin-bottom": "15px"}),
              ],
              style={"display": "inline-block", "margin-left": "20px"},
            ),
            html.H2("Display Settings"),
            html.Div(
              className="toggle-container",
              children=[
                dcc.Checklist(
                  id="toggle-legend",
                  options=[{"label": " Show Legend", "value": "show_legend"}],
                  value=[""],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
                dcc.Checklist(
                  id="toggle-title",
                  options=[{"label": " Show Title", "value": "show_title"}],
                  value=["show_title"],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
                dcc.Checklist(
                  id="toggle-lines",
                  options=[{"label": " Show Lines", "value": "show_lines"}],
                  value=[""],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
                dcc.Checklist(
                  id="toggle-labels",
                  options=[{"label": " Show Labels", "value": "show_labels"}],
                  value=["show_labels"],
                  labelStyle={"display": "block", "font-weight": "515", "margin-bottom": "5px"},
                ),
              ],
            ),
            html.H2("Export Settings"),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Download Format")]),
                dcc.Dropdown(
                  id="dl-format-dropdown",
                  options=[
                    {"label": "SVG", "value": "svg"},
                    {"label": "PNG", "value": "png"},
                    {"label": "JPEG", "value": "jpeg"},
                    {"label": "WEBP", "value": "webp"},
                  ],
                  value="svg",
                ),
              ],
              style={"margin-bottom": "5px"},
            ),
            html.Div(
              className="title-dropdown",
              children=[
                html.Div(className="dropdown-label", children=[html.Label("Background Color")]),
                dcc.Dropdown(
                  id="background-dropdown",
                  options=[
                    {"label": "Transparent", "value": "rgba(255, 255, 255, 0)"},
                    {"label": "White", "value": "rgba(255, 255, 255, 255)"},
                  ],
                  value="rgba(255, 255, 255, 0)",
                ),
              ],
            ),
          ],
        ),
        page_name=page_name,
      ),
      html.Div(
        id=f"content-{page_name}",
        children=[
          html.Div(
            [html.Div(id=f"graph-{page_name}", style={"width": "100%", "height": "100%"}, className="graph-container")],
            style={"width": "100%", "height": "100%"},
          ),
          html.Div(id="checklist-states", style={"display": "none"}),
        ],
        className="content",
        style={
          "marginLeft": navigation.side_bar_width,
          "width": "calc(100%-" + navigation.side_bar_width + ")",
          "height": "100%",
        },
      ),
    ],
    style={"width": "100%", "height": "100vh", "display": "flex", "flexDirection": "column"},
  )


def setup_callbacks(explorer):
  @explorer.app.callback(
    [
      Output("metric-x-dropdown", "options"),
      Output("metric-y-dropdown", "options"),
      Output(f"target-dropdown-{page_name}", "options"),
    ],
    Input("yaml-dropdown", "value"),
  )
  def update_dropdowns(selected_yaml):
    if not selected_yaml or selected_yaml not in explorer.dfs:
      return [], [], []

    df = explorer.dfs[selected_yaml]
    metrics_from_yaml = explorer.update_metrics(explorer.all_data[selected_yaml])
    available_metrics = [{"label": metric.replace("_", " "), "value": metric} for metric in metrics_from_yaml]
    available_targets = [{"label": target, "value": target} for target in df["Target"].unique()]

    return available_metrics, available_metrics, available_targets

  @explorer.app.callback(
    Output(f"graph-{page_name}", "children"),
    [
      Input("yaml-dropdown", "value"),
      Input("metric-x-dropdown", "value"),
      Input("metric-y-dropdown", "value"),
      Input(f"target-dropdown-{page_name}", "value"),
      Input("show-all", "n_clicks"),
      Input("hide-all", "n_clicks"),
      Input("toggle-legend", "value"),
      Input("toggle-title", "value"),
      Input("toggle-lines", "value"),
      Input("toggle-labels", "value"),
      Input("dl-format-dropdown", "value"),
      Input("background-dropdown", "value"),
    ]
    + [Input(f"checklist-{architecture}-{page_name}", "value") for architecture in explorer.all_architectures],
  )
  def update_graph(
    selected_yaml,
    selected_metric_x,
    selected_metric_y,
    selected_target,
    show_all,
    hide_all,
    toggle_legend,
    toggle_title,
    toggle_lines,
    toggle_labels,
    dl_format,
    background,
    *checklist_values,
  ):
    try:
      if not selected_yaml or selected_yaml not in explorer.dfs:
        return html.Div(className="error", children=[html.Div("Please select a YAML file.")])

      if not selected_target or selected_target not in explorer.dfs[selected_yaml]["Target"].values:
        return html.Div(className="error", children=[html.Div("Please select a valid target.")])

      selected_metric_x_display = selected_metric_x.replace("_", " ") if selected_metric_x is not None else ""
      selected_metric_y_display = selected_metric_y.replace("_", " ") if selected_metric_y is not None else ""

      unit_x = legend.unit_to_html(explorer.units[selected_yaml].get(selected_metric_x, ""))
      unit_y = legend.unit_to_html(explorer.units[selected_yaml].get(selected_metric_y, ""))

      selected_metric_x_display_unit = (
        selected_metric_x_display + " (" + unit_x + ")" if unit_x else selected_metric_x_display
      )
      selected_metric_y_display_unit = (
        selected_metric_y_display + " (" + unit_y + ")" if unit_y else selected_metric_y_display
      )

      ctx = dash.callback_context
      triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

      if triggered_id in ["show-all", "hide-all"]:
        visible_architectures = set(explorer.all_architectures if triggered_id == "show-all" else [])
      else:
        visible_architectures = set(
          architecture for i, architecture in enumerate(explorer.all_architectures) if checklist_values[i]
        )

      filtered_df = explorer.dfs[selected_yaml][
        (explorer.dfs[selected_yaml]["Target"] == selected_target)
        & (explorer.dfs[selected_yaml]["Architecture"].isin(visible_architectures))
      ]

      fig = go.Figure()
      for i, architecture in enumerate(explorer.all_architectures):
        if architecture in visible_architectures:
          df_architecture = filtered_df[filtered_df["Architecture"] == architecture]

          if selected_metric_x is None or selected_metric_x not in df_architecture.columns:
            return html.Div(className="error", children=[html.Div("Please select a valid x metric.")])
          if selected_metric_y is None or selected_metric_y not in df_architecture.columns:
            return html.Div(className="error", children=[html.Div("Please select a valid y metric.")])

          x_values = df_architecture[selected_metric_x].tolist()
          y_values = df_architecture[selected_metric_y].tolist()
          config_names = df_architecture["Configuration"].tolist()

          mode = "lines+markers" if "show_lines" in toggle_lines else "markers"
          if toggle_labels:
            mode += "+text"

          fig.add_trace(
            go.Scatter(
              x=x_values,
              y=y_values,
              mode=mode,
              line=dict(dash="dot") if "show_lines" in toggle_lines else None,
              marker=dict(size=10, color=legend.get_color(i)),
              name=architecture,
              connectgaps=True,
              text=config_names,
              textposition="top center",
              hovertemplate="<br>".join(
                [
                  "Architecture: %{fullData.name}",
                  "Configuration: %{text}",
                  selected_metric_x_display + ": %{x} " + unit_x,
                  selected_metric_y_display + ": %{y} " + unit_y,
                  "<extra></extra>",
                ]
              ),
            )
          )

      fig.update_layout(
        paper_bgcolor=background,
        showlegend="show_legend" in toggle_legend,
        xaxis_title=selected_metric_x_display_unit,
        yaxis_title=selected_metric_y_display_unit,
        xaxis=dict(range=[0, None]),
        yaxis=dict(range=[0, None]),
        title=selected_metric_y_display + " vs " + selected_metric_x_display if "show_title" in toggle_title else None,
        title_x=0.5,
        autosize=True,
      )
      filename = "Odatix-{}-{}-{}-vs-{}".format(
        os.path.splitext(selected_yaml)[0], selected_target, selected_metric_y, selected_metric_x
      )
      return html.Div(
        [
          dcc.Graph(
            figure=fig,
            style={"width": "100%", "height": "100%"},
            config={
              "displayModeBar": True,
              "displaylogo": False,
              "modeBarButtonsToRemove": ["lasso", "select"],
              "toImageButtonOptions": {
                "format": dl_format,
                "scale": "3",
                "filename": filename,
              },
            },
          )
        ],
        style={"width": "100%", "height": "100%", "display": "inline-block", "vertical-align": "top"},
      )
    except Exception as e:
      return html.Div(className="error", children=[html.Div("Unexpected error: " + str(e))])

  legend.setup_callbacks(explorer, page_name)
  navigation.setup_sidebar_callbacks(explorer, page_name)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import re
import queue
import fcntl
import curses
import select
import signal
import subprocess

from odatix.components.motd import read_version

from odatix.lib.ansi_to_curses import AnsiToCursesConverter
import odatix.lib.printc as printc

######################################
# Settings
######################################

STD_BUF = "stdbuf -oL "
script_name = os.path.basename(__file__)
error = None

######################################
# ParallelJob
######################################

class ParallelJob:
  status_file_pattern = re.compile(r"(.*)")
  progress_file_pattern = re.compile(r"(.*)")

  def __init__(
    self,
    process,
    command,
    directory,
    generate_rtl,
    generate_command,
    target,
    arch,
    display_name,
    status_file,
    progress_file,
    tmp_dir,
    progress_mode="default",
    status="not started",
  ):
    self.process = process
    self.command = command
    self.directory = directory
    self.generate_rtl = generate_rtl
    self.generate_command = generate_command
    self.target = target
    self.arch = arch
    self.display_name = display_name
    self.status_file = status_file
    self.progress_file = progress_file
    self.tmp_dir = tmp_dir
    self.progress_mode = progress_mode
    self.status = status

    self.log_history = []
    self.log_position = 0
    self.log_changed = False
    self.autoscroll = True

  @staticmethod
  def set_patterns(progress_file_pattern, status_file_pattern=None):
    ParallelJob.status_file_pattern = status_file_pattern
    ParallelJob.progress_file_pattern = progress_file_pattern

  def get_progress(self):
    if self.progress_mode == "fmax":
      return self.get_progress_fmax()
    else:
      progress = 0
      if os.path.isfile(self.progress_file):
        with open(self.progress_file, "r") as f:
          content = f.read()
        for match in re.finditer(ParallelJob.progress_file_pattern, content):
          parts = ParallelJob.progress_file_pattern.search(match.group())
          if len(parts.groups()) >= 2:
            progress = int(parts.group(2))
      if progress > 100:
        progress = 100
      return progress

  def get_progress_fmax(self):
    # Get progress from status file
    fmax_progress = 0
    fmax_step = 1
    fmax_totalstep = 1
    if os.path.isfile(self.status_file):
      with open(self.status_file, "r") as f:
        content = f.read()
      for match in re.finditer(ParallelJob.status_file_pattern, content):
        parts = ParallelJob.status_file_pattern.search(match.group())
        if len(parts.groups()) >= 4:
          fmax_progress = int(parts.group(2))
          fmax_step = int(parts.group(3))
          fmax_totalstep = int(parts.group(4))

    # Get progress from synth status file
    synth_progress = 0
    if os.path.isfile(self.progress_file):
      with open(self.progress_file, "r") as f:
        content = f.read()
      for match in re.finditer(ParallelJob.progress_file_pattern, content):
        parts = ParallelJob.progress_file_pattern.search(match.group())
        if len(parts.groups()) >= 2:
          synth_progress = int(parts.group(2))

    # Compute total progress
    if fmax_totalstep != 0:
      progress = fmax_progress + synth_progress / fmax_totalstep
    else:
      progress = synth_progress

    if progress > 100:
      progress = 100
    return progress


######################################
# ParallelJobHandler
######################################

class ParallelJobHandler:
  def __init__(self, job_list, nb_jobs=4, process_group=True, auto_exit=False, log_size_limit=100):
    self.job_list = job_list
    self.nb_jobs = nb_jobs
    self.process_group = process_group
    self.auto_exit = auto_exit
    self.log_size_limit = log_size_limit

    self.version = read_version()

    self.running_job_list = []
    self.retired_job_list = []
    self.job_queue = queue.Queue()
    self.selected_job_index = 0
    self.previous_log_size = 0
    self.max_title_length = max(len(job.display_name) for job in job_list)

    self.converter = AnsiToCursesConverter()

  @staticmethod
  def set_nonblocking(fd):
    try:
      fl = fcntl.fcntl(fd, fcntl.F_GETFL)
      fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
    except TypeError:
      pass

  @staticmethod
  def progress_bar(window, id, progress, bar_width, title, title_size, width, status="", selected=False):
    bar_width = bar_width - title_size
    if bar_width < 4:
      bar_width = 4
      title_size = width - bar_width - 25
      if title_size < 0:
        title_size = 0

      if len(title) > title_size:
        title = title[: title_size - 3] + "..."
      else:
        title = title.ljust(title_size)
    else:
      title = title.ljust(title_size)

    bar_length = int(bar_width * progress / 100.0)
    percentage = f"{progress:.0f}%"
    comment = f"({status})"

    try:
      button = "[*]" if selected else "[ ]"
      window.addstr(id, 0, f"{button} {title} [")
      window.addstr(id, len(button) + len(title) + 3, "#" * bar_length, curses.color_pair(1))
      window.addstr(id, len(button) + len(title) + 3 + bar_length, " " * (bar_width - bar_length), curses.color_pair(1))
      window.addstr(id, len(button) + len(title) + 3 + bar_width, f"] {percentage}")

      comment_position = len(button) + len(title) + 3 + bar_width + 8
      if status == "failed":
        window.addstr(id, comment_position, comment, curses.color_pair(2))
      elif status == "running":
        window.addstr(id, comment_position, comment, curses.color_pair(3))
      elif status == "success":
        window.addstr(id, comment_position, comment, curses.color_pair(4))
      elif status == "queued":
        window.addstr(id, comment_position, comment, curses.color_pair(5))
      elif status == "starting":
        window.addstr(id, comment_position, comment, curses.color_pair(6))
      else:
        window.addstr(id, comment_position, comment, curses.color_pair(1))
    except curses.error as e:
      pass

  def update_header(self, header_win, active_jobs_count, retired_jobs_count, total_jobs_count, width):
    try:
      header_win.hline(0, 0, " ", width, curses.color_pair(1) | curses.A_REVERSE)
      header_win.addstr(0, 1, "v" + str(self.version), curses.color_pair(1) | curses.A_REVERSE)
    except curses.error:
      pass
    if total_jobs_count == 1:
      if retired_jobs_count == total_jobs_count:
        text = "{}/{} job done!".format(retired_jobs_count, total_jobs_count)
      else:
        text = "{}/{} job done - {} running".format(retired_jobs_count, total_jobs_count, active_jobs_count)
    else:
      if retired_jobs_count == total_jobs_count:
        text = "{}/{} jobs done!".format(retired_jobs_count, total_jobs_count)
      else:
        text = "{}/{} jobs done - {} running".format(retired_jobs_count, total_jobs_count, active_jobs_count)
    try:
      header_win.addstr(0, width - len(text) - 1, text, curses.color_pair(1) | curses.A_REVERSE)
    except curses.error:
      pass

    try:
      header_win.addstr(
        0, (width - len(" Odatix ")) // 2, " Odatix ", curses.color_pair(1) | curses.A_REVERSE | curses.A_BOLD
      )
    except curses.error:
      pass
    try:
      header_win.hline(1, 0, "-", width)
    except curses.error:
      pass

    header_win.refresh()

  @staticmethod
  def update_help(help_win):
    help_win.erase()

    # Define the text with attributes
    help_text = [
      ("q", "Quit"),
      ("PageUp/PageDown", "Switch Process"),
      ("Up/Down", "Scroll Log"),
      ("Home/End", "Scroll to Top/Bottom"),
    ]

    help_win.attron(curses.color_pair(1) | curses.A_REVERSE)
    try:
      help_win.addstr(" ")
    except curses.error:
      pass

    for i, (key, description) in enumerate(help_text):
      try:
        if i > 0:
          help_win.addstr(" | ")
        help_win.attron(curses.color_pair(1) | curses.A_REVERSE | curses.A_BOLD)
        help_win.addstr(key)
        help_win.attroff(curses.A_BOLD)  # Remove attributes
        help_win.addstr(": ")
        help_win.addstr(description)
      except curses.error:
        pass

    try:
      help_win.addstr(" ")
    except curses.error:
      pass
    help_win.attroff(curses.color_pair(1) | curses.A_REVERSE | curses.A_BOLD)
    help_win.refresh()

  @staticmethod
  def show_exit_confirmation(help_win):
    try:
      help_win.erase()
      help_win.addstr(" Kill all jobs and exit: Yes (", curses.color_pair(1) | curses.A_REVERSE)
      help_win.addstr("y", curses.color_pair(1) | curses.A_REVERSE | curses.A_BOLD)
      help_win.addstr(") / No (", curses.color_pair(1) | curses.A_REVERSE)
      help_win.addstr("n", curses.color_pair(1) | curses.A_REVERSE | curses.A_BOLD)
      help_win.addstr(")? ", curses.color_pair(1) | curses.A_REVERSE)
      help_win.refresh()
    except curses.error:
      pass

    key = help_win.getch()
    curses.flushinp()
    if key == ord("y"):
      return True, True
    elif key == ord("n"):
      return True, False
    else:
      return False, False

  @staticmethod
  def update_exit(help_win):
    try:
      help_win.erase()
      help_win.addstr(" Exiting... ", curses.color_pair(1) | curses.A_REVERSE)
      help_win.refresh()
    except curses.error:
      pass

  def update_logs(self, logs_win, selected_job, logs_height, width):
    history = selected_job.log_history
    log_length = len(history)

    # Erase lines extra lines from previous selected job
    if log_length < self.previous_log_size:
      for i in range(log_length, self.previous_log_size):
        try:
          logs_win.move(i, 0)
          logs_win.clrtoeol()
        except curses.error:
          pass

    # Logs from selected job
    for i, line in enumerate(history[selected_job.log_position : selected_job.log_position + logs_height]):
      try:
        logs_win.move(i, 0)
        logs_win.clrtoeol()
        self.converter.add_ansi_str(logs_win, line, width=width)
      except curses.error:
        pass

    logs_win.refresh()
    self.previous_log_size = log_length

  def start_job(self, job):
    # Run generate command
    if not job.generate_rtl:
      self.run_job(job)
      self.running_job_list.append(job)
      return

    try:
      job.log_history.append(printc.colors.CYAN + "Run generate command for " + job.display_name + printc.colors.ENDC)
      job.log_history.append(printc.colors.BOLD + " > " + job.generate_command + printc.colors.ENDC)

      process = subprocess.Popen(
        job.generate_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=job.tmp_dir,
        shell=True,
        universal_newlines=True
      )

      self.set_nonblocking(process.stdout)
      self.set_nonblocking(process.stderr)

      job.process = process
      job.status = "starting"
      self.running_job_list.append(job)
    except subprocess.CalledProcessError:
      job.status = "failed"
      self.retire_job(job, progress=0)
      job.log_history.append(printc.colors.RED + "error: rtl generation failed" + printc.colors.ENDC)
      job.log_history.append(printc.colors.CYAN + "note: look for earlier error to solve this issue" + printc.colors.ENDC)
      return

  def run_job(self, job):
    job.log_history.append(printc.colors.CYAN + "Run job command" + printc.colors.ENDC)
    job.log_history.append(printc.colors.BOLD + " > " + job.command + printc.colors.ENDC)

    process = subprocess.Popen(
      STD_BUF + job.command,
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      cwd=job.directory,
      shell=True,
      universal_newlines=True,
      bufsize=1,
      preexec_fn=os.setpgrp if self.process_group else None, 
    )

    self.set_nonblocking(process.stdout)
    self.set_nonblocking(process.stderr)

    job.process = process
    job.status = "running"

  def queue_job(self, job):
    job.status = "queued"
    self.job_queue.put(job)

  def retire_job(self, job, progress=100):
    self.running_job_list.remove(job)
    job.progress = progress
    self.retired_job_list.append(job)

  def terminate_all_jobs(self):
    for job in self.running_job_list:
      if job.process:
        try:  # Try to terminate the process group
          os.killpg(os.getpgid(job.process.pid), signal.SIGTERM)
        except ProcessLookupError:
          pass  # Process already terminated

    # Wait for all processes to finish
    for job in self.running_job_list:
      if job.process:
        job.process.wait()

  def curses_main(self, stdscr):
    curses.curs_set(0)  # Hide cursor

    curses.start_color()
    curses.use_default_colors()

    # Define color pairs
    curses.init_pair(1, -1, -1)
    curses.init_pair(2, curses.COLOR_RED, -1)
    curses.init_pair(3, curses.COLOR_YELLOW, -1)
    curses.init_pair(4, curses.COLOR_GREEN, -1)
    curses.init_pair(5, curses.COLOR_BLUE, -1)
    curses.init_pair(6, curses.COLOR_CYAN, -1)
    curses.init_pair(10, curses.COLOR_BLACK, curses.COLOR_WHITE + AnsiToCursesConverter.LIGHT_OFFSET)

    height, width = stdscr.getmaxyx()
    old_width = width
    old_height = height

    header_height = 2

    # Adjust window positions
    progress_height = len(self.job_list)
    separator_height = 1
    help_height = 1
    logs_height = height - progress_height - separator_height - help_height - header_height

    try:
      header_win = curses.newwin(header_height, width, 0, 0)
      progress_win = curses.newwin(progress_height, width, header_height, 0)
      separator_win = curses.newwin(separator_height, width, header_height + progress_height, 0)
      logs_win = curses.newwin(logs_height, width, header_height + progress_height + separator_height, 0)
      help_win = curses.newwin(help_height, width, height - help_height, 0)
    except curses.error:
      stdscr.clear()
      try:
        stdscr.addstr(0, 0, "Could not start: window is too small. Press any key to exit", curses.color_pair(2))
        stdscr.refresh()
      except curses.error:
        pass
      stdscr.getch()
      sys.exit(-1)

    finished = False
    ask_exit = False

    selected_job = self.job_list[self.selected_job_index]

    for job in self.job_list:
      if len(self.running_job_list) < self.nb_jobs:
        self.start_job(job)
      else:
        self.queue_job(job)

    total_jobs_count = len(self.job_list)

    stdscr.nodelay(True)

    while True:
      height, width = stdscr.getmaxyx()
      # If window size changes, adjust the layout
      if height != old_height or width != old_width:
        try:
          logs_height = height - progress_height - separator_height - help_height - header_height
          logs_win = curses.newwin(logs_height, width, header_height + progress_height + separator_height, 0)
          help_win = curses.newwin(help_height, width, height - help_height, 0)
          progress_win = curses.newwin(progress_height, width, header_height, 0)
          separator_win = curses.newwin(separator_height, width, header_height + progress_height, 0)

          self.update_logs(logs_win, selected_job, logs_height, width)
        except curses.error:
          try:
            stdscr.clear()
            stdscr.addstr(0, 0, "Window is too small!", curses.color_pair(2))
            stdscr.refresh()
          except curses.error:
            pass

        old_width = width
        old_height = height

      # Check if all jobs have finished
      active_jobs_count = len(self.running_job_list)
      if active_jobs_count == 0:
        finished = True

      retired_jobs_count = len(self.retired_job_list)

      # Add a header
      self.update_header(header_win, active_jobs_count, retired_jobs_count, total_jobs_count, width)

      progress_win.erase()
      height, width = progress_win.getmaxyx()
      for id, job in enumerate(self.job_list):

        if job in self.retired_job_list:
          progress = job.progress
        elif job not in self.running_job_list or job.process is None:
          progress = 0
        else:
          progress = job.get_progress()
          if job.process.poll() is not None:
            if job.process.returncode == 0:
              if job.status == "starting":
                job.log_history.append("")
                self.run_job(job)
                continue
              else:
                job.status = "success"
            else:
              job.status = "failed"
              if progress is None:
                progress = 0

            self.retire_job(job, progress)
            if not self.job_queue.empty():
              self.start_job(self.job_queue.get())
              
            if job == selected_job:
              selected_job.log_changed = True
              self.update_logs(logs_win, selected_job, logs_height, width)
        try:
          self.progress_bar(
            id=id,
            window=progress_win,
            progress=progress,
            bar_width=(width - 25),
            title=job.display_name,
            title_size=self.max_title_length,
            width=width,
            status=job.status,
            selected=(id == self.selected_job_index),
          )
        except curses.error:
          pass
      progress_win.refresh()

      # Add a separator
      separator_win.erase()
      separator_win.hline(0, 0, "-", width)
      separator_win.refresh()

      # Collect all stdout and stderr pipes
      pipes = [job.process.stdout for job in self.running_job_list] + [
        job.process.stderr for job in self.running_job_list
      ]

      try:
        read_ready, _, _ = select.select(pipes, [], [], 0.1)
      except:
        read_ready = []

      for pipe in read_ready:
        for job in self.running_job_list:
          job.log_changed = False
          if pipe in (job.process.stdout, job.process.stderr):
            while True:
              line = pipe.readline()
              if line:
                job.log_history.append(line)
                # Apply log size limit
                if self.log_size_limit != -1 and len(job.log_history) > self.log_size_limit:
                  job.log_history = job.log_history[-self.log_size_limit :]
                job.log_changed = True
              else:
                break

      # Automatically scroll if at the bottom
      if selected_job.autoscroll and selected_job.log_changed:
        selected_job.log_position = max(0, len(selected_job.log_history) - logs_height)
        self.update_logs(logs_win, selected_job, logs_height, width)

      # Handle resize
      if width != old_width:
          self.update_logs(logs_win, selected_job, logs_height, width)
      old_width = width

      # Get inputs
      key = stdscr.getch()
      curses.flushinp()

      if key == curses.KEY_PPAGE:  # Page Up
        if self.selected_job_index > 0:
          self.selected_job_index -= 1
          selected_job = self.job_list[self.selected_job_index]
          selected_job.log_position = max(0, len(selected_job.log_history) - logs_height)
          selected_job.autoscroll = True
          self.update_logs(logs_win, selected_job, logs_height, width)
      elif key == curses.KEY_NPAGE:  # Page Down
        if self.selected_job_index < len(self.job_list) - 1:
          self.selected_job_index += 1
          selected_job = self.job_list[self.selected_job_index]
          selected_job.log_position = max(0, len(selected_job.log_history) - logs_height)
          selected_job.autoscroll = True
          self.update_logs(logs_win, selected_job, logs_height, width)
      elif key == curses.KEY_UP:  # Scroll Up
        if selected_job.log_position > 0:
          selected_job.log_position = max(0, selected_job.log_position - 3)
          selected_job.autoscroll = False
          self.update_logs(logs_win, selected_job, logs_height, width)
      elif key == curses.KEY_DOWN:  # Scroll Down
        if selected_job.log_position + logs_height < len(selected_job.log_history):
          selected_job.log_position = min(len(selected_job.log_history) - logs_height, selected_job.log_position + 3)
          selected_job.autoscroll = False
          self.update_logs(logs_win, selected_job, logs_height, width)
      elif key == curses.KEY_HOME:  # Home
        selected_job.log_position = 0
        selected_job.autoscroll = False
        self.update_logs(logs_win, selected_job, logs_height, width)
      elif key == curses.KEY_END:  # End
        selected_job.log_position = max(0, len(selected_job.log_history) - logs_height)
        selected_job.autoscroll = True
        self.update_logs(logs_win, selected_job, logs_height, width)
      else:
        if self.auto_exit and finished:
          return True
        else:
          if key == ord("q"):
            if finished:
              return True
            else:
              ask_exit = True

      # Handle exit
      if ask_exit:
        user_answered, user_confirmed = self.show_exit_confirmation(help_win)
        if user_answered:
          if user_confirmed:
            self.update_exit(help_win)
            self.terminate_all_jobs()
            return False
          else:
            ask_exit = False
      else:
        self.update_help(help_win)

  def run(self):
    curses.wrapper(self.curses_main)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import sys
import tty
import termios

def set_raw_mode():
  """
  Sets the terminal to raw mode, which allows reading input byte by byte without requiring Enter.
  This function saves the current terminal settings so they can be restored later.
  
  Returns:
    old_settings (list): The original terminal settings before switching to raw mode.
  """
  fd = sys.stdin.fileno()
  global old_settings
  old_settings = termios.tcgetattr(fd)
  tty.setraw(fd)
  return old_settings

def restore_mode(old_settings):
  """
  Restores the terminal to a previously saved state.

  Args:
    old_settings (list): The terminal settings to restore.
  """
  fd = sys.stdin.fileno()
  termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import yaml
import re
import csv
import argparse

import odatix.lib.printc as printc
from odatix.lib.utils import read_from_list, create_dir, KeyNotInListError, BadValueInListError
import odatix.lib.settings as settings
from odatix.lib.settings import OdatixSettings

current_dir = os.path.dirname(os.path.abspath(__file__))

banned_metrics = []
banned_arch = []

######################################
# Settings
######################################

DEFAULT_FORMAT = "yml"

simulations_dir = "simulations"

status_done = "Done: 100%"

script_name = os.path.basename(__file__)


######################################
# Parse Arguments
######################################


def add_arguments(parser):
  parser.add_argument("-t", "--tool", default="all", help="eda tool in use, or 'all' (default: all)")
  parser.add_argument(
    "-f",
    "--format",
    choices=["csv", "yml", "all"],
    default=DEFAULT_FORMAT,
    help="Output format: csv, yml, or all (default: " + DEFAULT_FORMAT + ")",
  )
  parser.add_argument("-u", "--use_benchmark", action="store_true", help="Use benchmark values in yaml file")
  parser.add_argument("-B", "--benchmark_file", help="Benchmark file")
  parser.add_argument("-w", "--work", help="Work directory")
  parser.add_argument("-r", "--respath", help="Result path")
  parser.add_argument(
    "-c",
    "--config",
    default=OdatixSettings.DEFAULT_SETTINGS_FILE,
    help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")",
  )


def parse_arguments():
  parser = argparse.ArgumentParser(description="Process FPGA or ASIC results")
  add_arguments(parser)
  return parser.parse_args()


######################################
# Parsing functions
######################################


def parse_regex(file, pattern, group_id, error_prefix=""):
  if not os.path.isfile(file):
    printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None
  with open(file, "r") as f:
    try:
      content = f.read()
      match = re.search(pattern, content)
      if match:
        return match.group(group_id)
    except Exception as e:
      printc.error(
        error_prefix + 'Could not get value from regex "' + pattern + '" in file "' + file + '": ' + str(e), script_name=script_name
      )
      return None

  printc.error(error_prefix + 'No match for regex "' + pattern + '" in file "' + file + '"', script_name=script_name)
  return None


def parse_csv(file, key, error_prefix=""):
  if not os.path.isfile(file):
    printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None
  with open(file, mode="r") as csv_file:
    try:
      reader = csv.DictReader(csv_file)
      for row in reader:
        if key in row:
          return row[key]
        else:
          printc.error(error_prefix + 'Could not find key "' + key + '" in csv "' + file + '"', script_name=script_name)
    except csv.Error as e:
      printc.error(error_prefix + 'An error occurred while reading csv file "' + file + '": ' + str(e), script_name=script_name)
      return None

  return None


def parse_yaml(file, key, error_prefix=""):
  if not os.path.isfile(file):
    printc.error(error_prefix + 'File "' + file + '" does not exist', script_name)
    return None

  with open(file, "r") as yaml_file:
    try:
      data = yaml.safe_load(yaml_file)
      keys = key.split("[")
      for k in keys:
        k = k.rstrip("]")
        if k in data:
          data = data[k]
        else:
          printc.error(error_prefix + 'Could not find key "' + k + '" in yaml "' + file + '"', script_name=script_name)
          return None
      return data
    except yaml.YAMLError as e:
      printc.error(error_prefix + 'Could not parse yaml file "' + file + '": ' + str(e), script_name=script_name)
      return None


######################################
# Validate Tool Settings
######################################


def validate_tool_settings(file_path):
  if not os.path.isfile(file_path):
    printc.error('Tool settings file "' + os.path.realpath(file_path) + '" does not exist', script_name)
    return None
  with open(file_path, "r") as file:
    try:
      tool_settings = yaml.safe_load(file)
      return tool_settings
    except yaml.YAMLError as e:
      printc.error("Error in tool configuration file: " + str(e), script_name)
      return None


######################################
# Extract Tool Metrics
######################################


def extract_metrics(tool_settings, tool_settings_file, cur_path, arch, arch_path, use_benchmark, benchmark_file):
  global banned_metrics
  results = {}
  units = {}
  error_prefix = arch_path + " => "
  metrics = read_from_list("metrics", tool_settings, tool_settings_file, raise_if_missing=False, script_name=script_name)
  for metric, content in metrics.items():
    if metric in banned_metrics:
      continue

    try:
      type = read_from_list("type", content, tool_settings_file, parent=metric, script_name=script_name)
      settings = read_from_list("settings", content, tool_settings_file, parent=metric, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      banned_metrics.append(metric)
      continue

    benchmark_only = read_from_list("benchmark_only", content, tool_settings_file, parent=metric, raise_if_missing=False, type=bool, print_error=False, script_name=script_name)
    if benchmark_only and not use_benchmark:
      banned_metrics.append(metric)
      continue

    if type == "regex":
      try:
        file = read_from_list("file", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
        pattern = read_from_list("pattern", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
        group_id = read_from_list("group_id", settings, tool_settings_file, parent=metric + "[settings]", type=int, script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = parse_regex(os.path.join(cur_path, file), pattern, group_id, error_prefix)
    elif type == "csv":
      try:
        file = read_from_list( "file", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
        key = read_from_list("key", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = parse_csv(os.path.join(cur_path, file), key, error_prefix)
    elif type == "yaml":
      try:
        file = read_from_list("file", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
        key = read_from_list("key", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = parse_yaml(os.path.join(cur_path, file), key, error_prefix)
    elif type == "benchmark":
      if not use_benchmark:
        banned_metrics.append(metric)
        continue
      if arch in banned_arch:
        continue
      try:
        key = read_from_list("key", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      key = arch + "[" + key + "]"
      value = parse_yaml(benchmark_file, key, error_prefix)
      if value is None:
        banned_arch.append(arch)
    elif type == "operation":
      try:
        op = read_from_list("op", settings, tool_settings_file, parent=metric + "[settings]", script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        banned_metrics.append(metric)
        continue
      value = calculate_operation(op, results, error_prefix)
    else:
      printc.error(
        'Unsupported metric type "' + type + '" specified for metric "' + metric + '" in "' + tool_settings_file + '"',
        script_name=script_name,
      )
      banned_metrics.append(metric)
      continue

    # Apply formatting if specified
    if value is not None and "format" in content:
      try:
        value = convert_to_numeric(content["format"] % float(value))
      except ValueError:
        pass  # printc.warning(f"Failed to format value {value} for metric {metric}", script_name)

    # Append unit if specified
    if value is not None:
      results[metric] = value
      if "unit" in content:
        units[metric] = content["unit"]
    else:
      results[metric] = None

  return results, units


######################################
# Misc functions
######################################


def corrupted_directory(directory):
  printc.warning(
    directory + " => Synthesis has not finished or directory has been corrupted", script_name
  )


def convert_to_numeric(data):
  try:
    if "." in data:
      return float(data)
    return int(data)
  except ValueError:
    return data


def calculate_operation(op_str, results, error_prefix=""):
  try:
    local_vars = {k: v for k, v in results.items() if v is not None}
    return eval(op_str, {}, local_vars)
  except (NameError, SyntaxError, TypeError, ZeroDivisionError) as e:
    printc.error(error_prefix + 'Failed to evaluate operation "' + op_str + '": ' + str(e) , script_name)
    return None


######################################
# Export Results
######################################


def export_results(input, output, tools, format, use_benchmark, benchmark_file):
  input_path = input
  for tool in tools:
    if tool == simulations_dir:
      continue
    
    printc.cyan("Export " + tool + " results", script_name)

    tool_settings_file = os.path.join(OdatixSettings.odatix_eda_tools_path, tool, "tool.yml")
    tool_settings = validate_tool_settings(tool_settings_file)
    if tool_settings is None:
      if len(tools) == 1:
        sys.exit(-1)
      else:
        continue

    data = {}
    units = {}

    input = os.path.join(input_path, tool)

    try:
      dirs = sorted(next(os.walk(input))[1])
    except StopIteration:
      continue

    for target in dirs:
      data[target] = {}
      for architecture in sorted(next(os.walk(os.path.join(input, target)))[1]):
        data[target][architecture] = {}
        for configuration in sorted(next(os.walk(os.path.join(input, target, architecture)))[1]):
          arch = architecture + "[" + configuration + "]"
          arch_path = os.path.join(target, architecture, configuration)
          cur_path = os.path.join(input, arch_path)

          status_log = os.path.join(cur_path, "log", "status.log")

          # Check if synthesis completed
          if not os.path.isfile(status_log):
            corrupted_directory(arch_path)
            continue

          with open(status_log, "r") as f:
            if status_done not in f.read():
              corrupted_directory(arch_path)
              continue

          # Get values
          metrics, cur_units = extract_metrics(tool_settings, tool_settings_file, cur_path, arch, arch_path, use_benchmark, benchmark_file)
          data[target][architecture][configuration] = metrics

          # Update units
          units.update(cur_units)

    # Export to the desired format
    os.makedirs(output, exist_ok=True)
    output_file = os.path.join(output, "results_" + tool + ".yml")
    try:
      with open(output_file, "w") as file:
        yaml.dump(
          {"units": units, "fmax_results": data}, file, default_style=None, default_flow_style=False, sort_keys=False
        )
        printc.say('Results written to "' + output_file + '"', script_name=script_name)
    except Exception as e:
      printc.error('Could not write "' + output_file + '"', script_name=script_name)
      printc.cyan("error details: ", script_name=script_name, end="")
      print(str(e))


######################################
# Main
######################################


def main(args, settings=None):
  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid:
      sys.exit(-1)

  if args.use_benchmark is not None:
    use_benchmark = args.use_benchmark
  else:
    use_benchmark = settings.use_benchmark

  if args.benchmark_file is not None:
    benchmark_file = args.benchmark_file
  else:
    benchmark_file = settings.benchmark_file

  if args.work is not None:
    input = args.work
  else:
    input = settings.fmax_work_path

  if not os.path.isdir(input):
    printc.error('Could not find fmax work directory "' + input + '"', script_name=script_name)
    printc.note("Run fmax synthesis using the 'odatix fmax' command before exporting the results", script_name=script_name)
    sys.exit(-1)

  if args.respath is not None:
    output = args.respath
  else:
    output = settings.result_path

  if args.tool == "all":
    tools = [item for item in os.listdir(input) if os.path.isdir(os.path.join(input, item))]
  else:
    tools = [args.tool]

  export_results(
    input=input,
    output=output,
    tools=tools,
    format=args.format,
    use_benchmark=use_benchmark,
    benchmark_file=benchmark_file,
  )


if __name__ == "__main__":
  args = parse_arguments()
  main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import yaml
import shutil
import argparse
import subprocess

import odatix.lib.printc as printc
from odatix.lib.replace_params import replace_params
from odatix.lib.parallel_job_handler import ParallelJobHandler, ParallelJob
from odatix.lib.settings import OdatixSettings
from odatix.lib.architecture_handler import ArchitectureHandler, Architecture
from odatix.lib.utils import read_from_list, copytree, create_dir, ask_to_continue, KeyNotInListError, BadValueInListError
from odatix.lib.prepare_work import edit_config_file
from odatix.lib.check_tool import check_tool
from odatix.lib.run_settings import get_synth_settings

current_dir = os.path.dirname(os.path.abspath(__file__))

######################################
# Settings
######################################

work_path = "work"

# Get eda_tools folder
if getattr(sys, "frozen", False):
  base_path = os.path.dirname(sys.executable)
else:
  base_path = current_dir
script_path = os.path.realpath(os.path.join(base_path, os.pardir, os.pardir, "odatix_eda_tools"))

work_script_path = "scripts"
work_report_path = "report"
work_result_path = "result"
work_log_path = "log"
common_script_path = "_common"
log_path = "log"
arch_path = "architectures"

nb_jobs = 4

param_settings_filename = "_settings.yml"
arch_filename = "architecture.txt"
target_filename = "target.txt"
tcl_config_filename = "settings.tcl"
yaml_config_filename = "settings.yml"
fmax_status_filename = "status.log"
synth_status_filename = "synth_status.log"
frequency_search_filename = "frequency_search.log"
tool_makefile_filename = "makefile.mk"
constraint_filename = "constraints.txt"
source_tcl = "source scripts/"
synth_fmax_rule = "synth_fmax_only"
test_tool_rule = "test_tool"

settings_ini_section = "SETTINGS"
valid_status = "Done: 100%"
valid_frequency_search = "Highest frequency with timing constraints being met"

progress_bar_size = 50
refresh_time = 5

default_fmax_lower_bound = 50
default_fmax_upper_bound = 500

fmax_status_pattern = re.compile(r"(.*): ([0-9]+)% \(([0-9]+)\/([0-9]+)\)(.*)")
synth_status_pattern = re.compile(r"(.*): ([0-9]+)%(.*)")

default_supported_tools = ["vivado", "design_compiler", "openlane"]

script_name = os.path.basename(__file__)


######################################
# Parse Arguments
######################################


def add_arguments(parser):
  parser.add_argument("-t", "--tool", default="vivado", help="eda tool in use (default: vivado)")
  parser.add_argument("-o", "--overwrite", action="store_true", help="overwrite existing results")
  parser.add_argument("-y", "--noask", action="store_true", help="do not ask to continue")
  parser.add_argument("-i", "--input", help="input settings file")
  parser.add_argument("-a", "--archpath", help="architecture directory")
  parser.add_argument("-w", "--work", help="work directory")
  parser.add_argument(
    "-c",
    "--config",
    default=OdatixSettings.DEFAULT_SETTINGS_FILE,
    help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")",
  )


def parse_arguments():
  parser = argparse.ArgumentParser(description="Run fmax synthesis on selected architectures")
  add_arguments(parser)
  return parser.parse_args()


######################################
# Run Synthesis
######################################


def run_synthesis(run_config_settings_filename, arch_path, tool, work_path, target_path, overwrite, noask):
  _overwrite, ask_continue, show_log_if_one, nb_jobs, architectures = get_synth_settings(run_config_settings_filename)

  work_path = os.path.join(work_path, tool)

  if architectures is None:
    printc.error('The "architectures" section of "' + run_config_settings_filename + '" is empty.', script_name)
    printc.note('You must define your architectures in "' + run_config_settings_filename + '" before using this command.', script_name)
    printc.note("Check out examples Odatix's documentation for more information.", script_name)
    sys.exit(-1)

  if overwrite:
    overwrite = True
  else:
    overwrite = _overwrite

  if noask:
    ask_continue = False

  eda_target_filename = os.path.realpath(os.path.join(target_path, "target_" + tool + ".yml"))

  # Check if the target file exists
  if not os.path.isfile(eda_target_filename):
    printc.error(
      'Target file "' + eda_target_filename + '", for the selected eda tool "' + tool + '" does not exist', script_name
    )
    sys.exit(-1)

  # Check if the tool has a dedicated directory in script_path
  eda_tool_dir = script_path + "/" + tool
  if not os.path.isdir(eda_tool_dir):
    printc.error(
      'The directory "' + eda_tool_dir + '", for the selected eda tool "' + tool + '" does not exist', script_name
    )
    if tool not in default_supported_tools:
      printc.note(
        'The selected eda tool "'
        + tool
        + "\" is not one of the supported tool. Check out Odatix's documentation to add support for your own eda tool",
        script_name,
      )
    sys.exit(-1)

  # Check if the tool makefile exists
  tool_makefile_file = script_path + "/" + tool + "/" + tool_makefile_filename
  if not os.path.isfile(tool_makefile_file):
    printc.error(
      'Makefile "' + tool_makefile_file + '", for the selected eda tool "' + tool + '" does not exist', script_name
    )
    if tool not in default_supported_tools:
      printc.note(
        'The selected eda tool "'
        + tool
        + "\" is not one of the supported tool. Check out Odatix's documentation to add support for your own eda tool",
        script_name,
      )
    sys.exit(-1)

  # Get tool settings
  tool_settings_filename = os.path.realpath(os.path.join(eda_tool_dir, "tool.yml"))

  # Check if the tool file exists 
  if not os.path.isfile(tool_settings_filename):
    printc.error(
      'Settings file "' + tool_settings_filename + '", for the selected eda tool "' + tool + '" does not exist', script_name
    )
    sys.exit(-1)
  with open(tool_settings_filename, "r") as f:
    try:
      settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    except Exception as e:
      printc.error('Settings file "' + tool_settings_filename + '", for the selected eda tool "' + tool + '" is not a valid YAML file', script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)

    # Mandatory keys
    try:
      process_group = read_from_list("process_group", settings_data, tool_settings_filename, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      pass

    global work_report_path
    try:
      work_report_path = read_from_list("report_path", settings_data, tool_settings_filename, optional=True, print_error=False, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      pass

  with open(eda_target_filename, "r") as f:
    try:
      settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    except Exception as e:
      printc.error('Settings file "' + eda_target_filename + '" is not a valid YAML file', script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)

    # Mandatory keys
    try:
      targets = read_from_list("targets", settings_data, eda_target_filename, script_name=script_name)
      constraint_file = read_from_list("constraint_file", settings_data, eda_target_filename, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      sys.exit(-1)  # if a key is missing

    # Optional keys
    try:
      install_path = read_from_list("tool_install_path", settings_data, eda_target_filename, print_error=False, script_name=script_name)
      install_path = os.path.realpath(os.path.expanduser(install_path))
      if not os.path.isdir(install_path):
        printc.error('The installation path "' + install_path + '" defined for tool "' + tool + '" in "' + eda_target_filename + '" does not exist', script_name)
        printc.note('Please update the path in "' + eda_target_filename + '"', script_name=script_name)
        printc.note('if no installation path is needed by ' + tool + '\'s Makefile, simply remove "install_path" from "' + eda_target_filename + '"', script_name=script_name)
        sys.exit(-1)

    except (KeyNotInListError, BadValueInListError):
      printc.note('No tool_install_path specified for "' + tool + '"', script_name=script_name)
      install_path = "/"

  # Try launching eda tool
  check_tool(
    tool, script_path, makefile=tool_makefile_filename, rule=test_tool_rule, supported_tools=default_supported_tools, tool_install_path=install_path
  )

  ParallelJob.set_patterns(synth_status_pattern, fmax_status_pattern)

  arch_handler = ArchitectureHandler(
    work_path=work_path,
    arch_path=arch_path,
    script_path=script_path,
    work_script_path=work_script_path,
    work_report_path=work_report_path,
    log_path=log_path,
    process_group=process_group,
    eda_target_filename=eda_target_filename,
    fmax_status_filename=fmax_status_filename,
    frequency_search_filename=frequency_search_filename,
    param_settings_filename=param_settings_filename,
    valid_status=valid_status,
    valid_frequency_search=valid_frequency_search,
    default_fmax_lower_bound=default_fmax_lower_bound,
    default_fmax_upper_bound=default_fmax_upper_bound,
    overwrite=overwrite,
  )

  architecture_instances = arch_handler.get_architectures(architectures, targets, constraint_file, install_path)

  # Print checklist summary
  arch_handler.print_summary()

  if arch_handler.get_valid_arch_count() > 0:
    if ask_continue:
      print()
      ask_to_continue()
  else:
    sys.exit(-1)

  print()

  job_list = []

  def prepare_job(arch_instance):
    if True:
      # Get param dir (arch name before '/')
      arch_param_dir = re.sub("/.*", "", arch_instance.arch_name)

      # Create directory
      create_dir(arch_instance.tmp_dir)

      # Copy scripts
      try:
        copytree(script_path + "/" + common_script_path, arch_instance.tmp_script_path)
      except:
        printc.error('"' + arch_instance.tmp_script_path + '" exists while it should not', script_name)

      copytree(script_path + "/" + tool + "/tcl", arch_instance.tmp_script_path, dirs_exist_ok=True)

      # Copy design
      if arch_instance.design_path != -1:
        copytree(arch_instance.design_path, arch_instance.tmp_dir, dirs_exist_ok=True)

      # Copy rtl (if exists)
      if not arch_instance.generate_rtl:
        copytree(arch_instance.rtl_path, arch_instance.tmp_dir + "/" + "rtl", dirs_exist_ok=True)

      # Replace parameters
      if arch_instance.use_parameters:
        # printc.subheader("Replace parameters")
        param_target_file = arch_instance.tmp_dir + "/" + arch_instance.param_target_filename
        param_filename = arch_path + "/" + arch_instance.arch_name + ".txt"
        replace_params(
          base_text_file=param_target_file,
          replacement_text_file=param_filename,
          output_file=param_target_file,
          start_delimiter=arch_instance.start_delimiter,
          stop_delimiter=arch_instance.stop_delimiter,
          replace_all_occurrences=False,
          silent=True,
        )
        # print()

      # Create target and architecture files
      f = open(arch_instance.tmp_dir + "/" + target_filename, "w")
      print(arch_instance.target, file=f)
      f.close()
      f = open(arch_instance.tmp_dir + "/" + arch_filename, "w")
      print(arch_instance.arch_name, file=f)
      f.close()

      # File copy
      if arch_instance.file_copy_enable:
        file_copy_dest = os.path.join(arch_instance.tmp_dir, arch_instance.file_copy_dest)
        try:
          shutil.copy2(arch_instance.file_copy_source, file_copy_dest)
        except Exception as e:
          printc.error(
            'Could not copy "' + arch_instance.script_copy_source + '" to "' + os.path.realpath(file_copy_dest) + '"',
            script_name,
          )
          printc.cyan("error details: ", end="", script_name=script_name)
          print(str(e))
          return

      # Script copy
      if arch_instance.script_copy_enable:
        try:
          shutil.copy2(arch_instance.script_copy_source, arch_instance.tmp_script_path)
        except Exception as e:
          printc.error(
            'Could not copy "'
            + arch_instance.script_copy_source
            + '" to "'
            + os.path.realpath(arch_instance.tmp_script_path)
            + '"',
            script_name,
          )
          printc.cyan("error details: ", end="", script_name=script_name)
          print(str(e))
          return

      # Edit tcl config script
      tcl_config_file = os.path.join(arch_instance.tmp_script_path, tcl_config_filename)
      report_path = os.path.join(arch_instance.tmp_dir, work_report_path)
      edit_config_file(arch_instance, tcl_config_file)

      # Write yaml config script
      yaml_config_file = os.path.join(arch_instance.tmp_dir, yaml_config_filename)
      Architecture.write_yaml(arch_instance, yaml_config_file)

      # Link all scripts to config script
      for filename in os.listdir(arch_instance.tmp_script_path):
        if filename.endswith(".tcl"):
          with open(arch_instance.tmp_script_path + "/" + filename, "r") as f:
            tcl_content = f.read()
          pattern = re.escape(source_tcl) + r"(.+?\.tcl)"

          def replace_path(match):
            return "source " + os.path.realpath(arch_instance.tmp_script_path) + "/" + match.group(1)

          tcl_content = re.sub(pattern, replace_path, tcl_content)
          with open(arch_instance.tmp_script_path + "/" + filename, "w") as f:
            f.write(tcl_content)

      # Run binary search script
      tool_makefile_file = script_path + "/" + tool + "/" + tool_makefile_filename
      command = (
        "make -f {} {}".format(tool_makefile_file, synth_fmax_rule)
        + ' WORK_DIR="{}"'.format(os.path.realpath(arch_instance.tmp_dir))
        + ' TOOL_INSTALL_PATH="{}"'.format(os.path.realpath(arch_instance.install_path))
        + ' ODATIX_DIR="{}"'.format(OdatixSettings.odatix_path)
        + ' SCRIPT_DIR="{}"'.format(os.path.realpath(os.path.join(arch_instance.tmp_dir, work_script_path)))
        + ' LOG_DIR="{}"'.format(os.path.realpath(os.path.join(arch_instance.tmp_dir, log_path)))
        + ' CLOCK_SIGNAL="{}"'.format(arch_instance.clock_signal)
        + ' TOP_LEVEL_MODULE="{}"'.format(arch_instance.top_level_module)
        + ' LIB_NAME="{}"'.format(arch_instance.lib_name)
        + " --no-print-directory"
      )

      fmax_status_file = os.path.join(arch_instance.tmp_dir, log_path, fmax_status_filename)
      synth_status_file = os.path.join(arch_instance.tmp_dir, log_path, synth_status_filename)

      running_arch = ParallelJob(
        process=None,
        command=command,
        directory=".",
        generate_rtl=arch_instance.generate_rtl,
        generate_command=arch_instance.generate_command,
        target=arch_instance.target,
        arch=arch_instance.arch_name,
        display_name=arch_instance.arch_display_name,
        status_file=fmax_status_file,
        progress_file=synth_status_file,
        tmp_dir=arch_instance.tmp_dir,
        progress_mode="fmax",
        status="idle",
      )

      job_list.append(running_arch)

  for arch_instance in architecture_instances:
    prepare_job(arch_instance)

  parallel_jobs = ParallelJobHandler(job_list, nb_jobs, arch_handler.process_group)
  job_exit_success = parallel_jobs.run()

  # Summary
  if job_exit_success:
    print()
    for running_arch in job_list:
      tmp_dir = work_path + "/" + running_arch.target + "/" + running_arch.arch
      frequency_search_file = tmp_dir + "/" + log_path + "/" + frequency_search_filename
      try:
        with open(frequency_search_file, "r") as file:
          lines = file.readlines()
          if len(lines) >= 1:
            summary_line = lines[-1]
            print(running_arch.display_name + ": " + summary_line, end="")
      except:
        pass
    print()


######################################
# Main
######################################


def main(args, settings=None):
  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid:
      sys.exit(-1)

  if args.input is not None:
    run_config_settings_filename = args.input
  else:
    run_config_settings_filename = settings.fmax_synthesis_settings_file

  if args.archpath is not None:
    arch_path = args.archpath
  else:
    arch_path = settings.arch_path

  if args.work is not None:
    work_path = args.work
  else:
    work_path = settings.fmax_work_path

  target_path = settings.target_path
  tool = args.tool
  overwrite = args.overwrite
  noask = args.noask

  run_synthesis(run_config_settings_filename, arch_path, tool, work_path, target_path, overwrite, noask)


if __name__ == "__main__":
  args = parse_arguments()
  main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import argparse
from pathlib import Path

# Add parent dir to path
current_dir = os.path.dirname(os.path.abspath(__file__))
sources_dir = os.path.realpath(os.path.join(current_dir, os.pardir))
sys.path.append(sources_dir)

from odatix.components.motd import *
import odatix.components.run_simulations as run_sim
import odatix.components.run_fmax_synthesis as run_synth
import odatix.components.export_results as exp_res
import odatix.components.export_benchmark as exp_bench
import odatix.components.clean as cln

import odatix.lib.settings as settings
from odatix.lib.settings import OdatixSettings
import odatix.lib.printc as printc
from odatix.lib.utils import *

######################################
# Settings
######################################

EXIT_SUCCESS = 0

error_logfile = "odatix_error.log"

prog = os.path.basename(sys.argv[0])
script_name = os.path.basename(__file__)

######################################
# Parse Arguments
######################################

class ArgParser:

  def add_nobanner(parser):
    parser.add_argument('-Q', '--nobanner', action='store_true', help='suppress printing of banner')

  def parse_arguments():
    formatter = OdatixHelpFormatter
    ArgParser.parser = argparse.ArgumentParser(
      conflict_handler="resolve", 
      formatter_class=formatter,
      add_help=False
    )
    ArgParser.parser.add_argument('-v', '--version', action='store_true', help='show version and exit')
    ArgParser.parser.add_argument('-h', '--help', action='store_true', help='show this help message and exit')
    ArgParser.parser.add_argument('--init', action='store_true', help="init current directory by adding all the necessary config files")
    args, remaining_args = ArgParser.parser.parse_known_args()

    # Parse other arguments
    subparsers = ArgParser.parser.add_subparsers(dest="command")
    
    # Define parser for the 'init' command
    ArgParser.init_parser = subparsers.add_parser("init", help="init current directory by adding all the necessary config files (without any prompt)", formatter_class=formatter)
    ArgParser.init_parser.add_argument("-e", "--examples", action="store_true", help="Include examples")

    # Define parser for the 'fmax' command
    ArgParser.fmax_parser = subparsers.add_parser("fmax", help="run fmax synthesis", formatter_class=formatter)
    run_synth.add_arguments(ArgParser.fmax_parser)
    ArgParser.fmax_parser.add_argument('-e', '--noexport', action='store_true', help='do not export results after synthesis')
    ArgParser.add_nobanner(ArgParser.fmax_parser)

    # Define parser for the 'sim' command
    ArgParser.sim_parser = subparsers.add_parser("sim", help="run simulations", formatter_class=formatter)
    run_sim.add_arguments(ArgParser.sim_parser)
    ArgParser.add_nobanner(ArgParser.sim_parser)

    # Define parser for the 'results' command
    ArgParser.res_parser = subparsers.add_parser("results", help="export benchmark results", formatter_class=formatter)
    ArgParser.res_parser.add_argument("-t", "--tool", default="all", help="eda tool in use, or 'all'")
    ArgParser.res_parser.add_argument("-f", "--format", choices=["csv", "yml", "all"], help="Output format: csv, yml, or all")
    ArgParser.res_parser.add_argument("-u", "--use_benchmark", action="store_true", help="Use benchmark values in yaml file")
    ArgParser.res_parser.add_argument('-b', '--benchmark', choices=['dhrystone'], default=exp_bench.DEFAULT_BENCHMARK, help='benchmark to parse (default: ' + exp_bench.DEFAULT_BENCHMARK + ')')
    ArgParser.res_parser.add_argument("-B", "--benchmark_file", help="output benchmark file")
    ArgParser.res_parser.add_argument('-S', '--sim_file', default=exp_bench.DEFAULT_SIM_FILE, help='simulation log file (default: ' + exp_bench.DEFAULT_SIM_FILE + ')')
    ArgParser.res_parser.add_argument("-w", "--work", help="simulation work directory")
    ArgParser.res_parser.add_argument("-r", "--respath", help="Result path")
    ArgParser.res_parser.add_argument("-c", "--config", default=OdatixSettings.DEFAULT_SETTINGS_FILE, help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")")
    ArgParser.add_nobanner(ArgParser.res_parser)

    # Define parser for the 'res_benchmark' command
    ArgParser.bm_res_parser = subparsers.add_parser("res_benchmark", help="export benchmark results", formatter_class=formatter)
    exp_bench.add_arguments(ArgParser.bm_res_parser)
    ArgParser.add_nobanner(ArgParser.bm_res_parser)

    # Define parser for the 'res_synth' command
    ArgParser.exp_res_parser = subparsers.add_parser("res_synth", help="export synthesis results")
    exp_res.add_arguments(ArgParser.exp_res_parser)
    ArgParser.add_nobanner(ArgParser.exp_res_parser)

    # Define parser for the 'clean' command
    ArgParser.clean_parser = subparsers.add_parser("clean", help="clean directory", formatter_class=formatter)
    cln.add_arguments(ArgParser.clean_parser)
    ArgParser.add_nobanner(ArgParser.clean_parser)

    # Parse arguments
    return ArgParser.parser.parse_args()

  def help():
    full_header(description="Odatix - a FPGA/ASIC toolbox for design space exploration")
    printc.bold("Global:\n  ", printc.colors.CYAN, end="")
    ArgParser.parser.print_help()
    print()
    printc.bold("Synthesis:\n  ", printc.colors.CYAN, end="")
    ArgParser.fmax_parser.print_help()
    print()
    printc.bold("Simulation:\n  ", printc.colors.CYAN, end="")
    ArgParser.sim_parser.print_help()
    print()
    printc.bold("Results:", printc.colors.CYAN)
    printc.cyan("- All Results:\n  ", end="")
    ArgParser.res_parser.print_help()
    print()
    printc.cyan("- Benchmark Results:\n  ", end="")
    print(ArgParser.bm_res_parser.format_usage(), end="")
    print("  run ", end="")
    printc.bold(prog + " res_benchmark -h", end="")
    print(" for more details")
    print()
    printc.cyan("- Synthesis Results:\n  ", end="")
    print(ArgParser.exp_res_parser.format_usage(), end="")
    print("  run ", end="")
    printc.bold(prog + " res_synth -h", end="")
    print(" for more details")
    print()
    printc.bold("Clean:\n  ", printc.colors.CYAN, end="")
    ArgParser.clean_parser.print_help()
    print()
    printc.bold("Init:\n  ", printc.colors.CYAN, end="")
    ArgParser.init_parser.print_help()
    print()

class OdatixHelpFormatter(argparse.HelpFormatter):
  def __init__(self, prog):
    super().__init__(
      prog=prog,
      max_help_position=8
    )
    self._current_indent = 2

  def _format_action_invocation(self, action):
    parts = super()._format_action_invocation(action).split(", ")
    parts = [printc.colors.BOLD + part + printc.colors.ENDC for part in parts]
    return ", ".join(parts)

######################################
# Run functions
######################################

def run_simulations(args):
  success = True
  try:
    run_sim.main(args)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False
  return success

def run_fmax_synthesis(args):
  success = True
  try:
    run_synth.main(args)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False
  if success and not args.noexport:
    try:
      newargs = argparse.Namespace(
        tool = args.tool,
        format = exp_res.DEFAULT_FORMAT,
        use_benchmark = None,
        benchmark_file = None,
        work = args.work,
        respath = None,
        config = args.config,
      )
      exp_res.main(newargs)
    except SystemExit as e:
      if e.code != EXIT_SUCCESS:
        success = False
    except Exception as e:
      internal_error(e, error_logfile, script_name)
      success = False
  return success

def export_benchmark(args):
  success = True
  try:
    exp_bench.main(args)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False

def export_results(args):
  success = True
  try:
    exp_res.main(args)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False
  return success

def export_all_results(args):
  success = True
  try:
    if args.use_benchmark:
      newargs = argparse.Namespace(
        benchmark = args.benchmark,
        sim_file = args.sim_file,
        benchmark_file = args.benchmark_file,
        work = args.work,
        config = args.config,
      )
      exp_bench.main(newargs)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
      return success
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False
    return success
  try:
    newargs = argparse.Namespace(
      tool = args.tool,
      format = args.format,
      use_benchmark = args.use_benchmark,
      benchmark_file = args.benchmark_file,
      work = args.work,
      respath = args.respath,
      config = args.config,
    )
    exp_res.main(newargs)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False
  return success

def clean(args):
  success = True
  try:
    cln.main(args)
  except SystemExit as e:
    if e.code != EXIT_SUCCESS:
      success = False
  except Exception as e:
    internal_error(e, error_logfile, script_name)
    success = False


######################################
# Main
######################################

def main(args=None):

  if args is None:
    args = ArgParser.parse_arguments()

  # Display the version if requested
  if args.version:
    print_version()
    sys.exit(0)

  # Display help if requested
  if args.help:
    ArgParser.help()
    sys.exit(0)

  # If no command is selected 
  if args.command is None and not args.init:
    full_header()
    print("run ", end="")
    printc.bold(prog + " -h", end="")
    print(" to get a list of useful commands")
    sys.exit(0)

  try:
    args.nobanner
  except AttributeError:
    args.nobanner = False

  # Display init dialog
  if args.init:
    OdatixSettings.init_directory_dialog(prog=prog)
    sys.exit(0)

  # Display the banner
  if not args.nobanner:
    motd()
    print()

  # Dispatch the command to the appropriate function
  if args.command == "init":
    success = OdatixSettings.init_directory_nodialog(args.examples, prog)
  elif args.command == "sim":
    success = run_simulations(args)
  elif args.command == "fmax":
    success = run_fmax_synthesis(args)
  elif args.command == "results":
    success = export_all_results(args)
  elif args.command in "res_benchmark":
    success = export_benchmark(args)
  elif args.command in "res_synth":
    success = export_results(args)
  elif args.command in "clean":
    success = clean(args)

  sys.exit(success)

if __name__ == "__main__":
  main()



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import yaml

import odatix.lib.printc as printc
from odatix.lib.utils import read_from_list, KeyNotInListError, BadValueInListError, ask_yes_no, YAML_BOOL, copytree

current_dir = os.path.dirname(os.path.abspath(__file__))
script_name = os.path.basename(__file__)

######################################
# Settings
######################################

# get eda_tools folder
if getattr(sys, "frozen", False):
  base_path = os.path.dirname(sys.executable)
else:
  base_path = current_dir

class OdatixSettings:
  DEFAULT_SETTINGS_FILE = "odatix.yml"

  DEFAULT_WORK_PATH = "work"
  DEFAULT_SIM_WORK_PATH = "work/simulations"
  DEFAULT_FMAX_WORK_PATH = "work/fmax"
  DEFAULT_RESULT_PATH = "results"
  DEFAULT_USERCONFIG_PATH = "odatix_userconfig"
  DEFAULT_ARCH_PATH = os.path.join(DEFAULT_USERCONFIG_PATH, "architectures")
  DEFAULT_SIM_PATH = os.path.join(DEFAULT_USERCONFIG_PATH, "simulations")
  DEFAULT_TARGET_PATH = DEFAULT_USERCONFIG_PATH
  DEFAULT_USE_BENCHMARK = False
  DEFAULT_BENCHMARK_FILE = "results/benchmark.yml"
  DEFAULT_CLEAN_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "clean.yml")
  DEFAULT_SIMULATION_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "simulations_settings.yml")
  DEFAULT_FMAX_SYNTHESIS_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "fmax_synthesis_settings.yml")
  # DEFAULT_RANGE_SYNTHESIS_SETTINGS_FILE = os.path.join(DEFAULT_USERCONFIG_PATH, "range_synthesis_settings.yml")
  
  odatix_path = os.path.realpath(os.path.join(base_path, os.pardir))
  odatix_eda_tools_path = os.path.realpath(os.path.join(odatix_path, os.pardir, "odatix_eda_tools"))
  odatix_init_path = os.path.realpath(os.path.join(odatix_path, os.pardir, "odatix_init"))
  odatix_examples_path = os.path.realpath(os.path.join(odatix_path, os.pardir, "odatix_examples"))

  def __init__(self, settings_filename=DEFAULT_SETTINGS_FILE):
    self.settings_file_exists = os.path.isfile(settings_filename)
    if self.settings_file_exists:
      success = self.read_settings_file(settings_filename)
    else:
      printc.warning("This directory does not contain an Odatix settings file \"" + settings_filename + "\".", script_name=script_name)
      printc.note("Run Odatix with the '--init' flag to generate it alongside other necessary config files", script_name=script_name)
      self.valid = False

  def read_settings_file(self, settings_filename=DEFAULT_SETTINGS_FILE):
    if not os.path.isfile(settings_filename):
      printc.error("Odatix settings file \"" + settings_filename + "\" does not exists.", script_name=script_name)
      printc.note("Odatix settings file should be in \"" + current_dir + "\"", script_name=script_name)
      self.valid = False
      return False

    with open(settings_filename, "r") as f:
      try:
        settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
      except Exception as e:
        printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
        printc.cyan("error details: ", end="", script_name=script_name)
        print(str(e))
        self.valid = False
        return False
      try:
        self.work_path = read_from_list("work_path", settings_data, settings_filename, script_name=script_name)
        self.sim_work_path = read_from_list("sim_work_path", settings_data, settings_filename, script_name=script_name)
        self.fmax_work_path = read_from_list("fmax_work_path", settings_data, settings_filename, script_name=script_name)
        self.result_path = read_from_list("result_path", settings_data, settings_filename, script_name=script_name)
        self.arch_path = read_from_list("arch_path", settings_data, settings_filename, script_name=script_name)
        self.sim_path = read_from_list("sim_path", settings_data, settings_filename, script_name=script_name)
        self.target_path = read_from_list("target_path", settings_data, settings_filename, script_name=script_name)
        self.use_benchmark = read_from_list("use_benchmark", settings_data, settings_filename, type=bool, script_name=script_name)
        self.benchmark_file = read_from_list("benchmark_file", settings_data, settings_filename , script_name=script_name)
        self.clean_settings_file = read_from_list("clean_settings_file", settings_data, settings_filename , script_name=script_name)
        self.simulation_settings_file = read_from_list("simulation_settings_file", settings_data, settings_filename , script_name=script_name)
        self.fmax_synthesis_settings_file = read_from_list("fmax_synthesis_settings_file", settings_data, settings_filename , script_name=script_name)
        # self.range_synthesis_settings_file = read_from_list("range_synthesis_settings_file", settings_data, settings_filename , script_name=script_name)
      except (KeyNotInListError, BadValueInListError):
        self.valid = False
        return False
    self.valid = True
    return True
    
  def generate_settings_file(self, settings_filename=DEFAULT_SETTINGS_FILE):
    if os.path.isfile(settings_filename):
      return True
    else:
      printc.say("This directory does not contain an Odatix settings file \"" + settings_filename + "\".", script_name=script_name)
      printc.say("This file is mandatory. Would you like to create one? ", end="", script_name=script_name)
      answer = ask_yes_no()
      if answer is False:
        return False
      else:
        settings_data = {
          "work_path": input("  Enter work path [default: " + OdatixSettings.DEFAULT_WORK_PATH + "]: ") or OdatixSettings.DEFAULT_WORK_PATH,
          "sim_work_path": input("  Enter simulation work path [default: " + OdatixSettings.DEFAULT_SIM_WORK_PATH + "]: ") or OdatixSettings.DEFAULT_SIM_WORK_PATH,
          "fmax_work_path": input("  Enter simulation work path [default: " + OdatixSettings.DEFAULT_FMAX_WORK_PATH + "]: ") or OdatixSettings.DEFAULT_FMAX_WORK_PATH,
          "result_path": input("  Enter result path [default: " + OdatixSettings.DEFAULT_RESULT_PATH + "]: ") or OdatixSettings.DEFAULT_RESULT_PATH,
          "arch_path": input("  Enter architecture path [default: " + OdatixSettings.DEFAULT_ARCH_PATH + "]: ") or OdatixSettings.DEFAULT_ARCH_PATH,
          "sim_path": input("  Enter simulation path [default: " + OdatixSettings.DEFAULT_SIM_PATH + "]: ") or OdatixSettings.DEFAULT_SIM_PATH,
          "target_path": input("  Enter target settings files path [default: " + OdatixSettings.DEFAULT_TARGET_PATH + "]: ") or OdatixSettings.DEFAULT_TARGET_PATH,
          "use_benchmark": input("  Use benchmark (True/False) [default: " + str(OdatixSettings.DEFAULT_USE_BENCHMARK) + "]: ").lower() in YAML_BOOL or OdatixSettings.DEFAULT_USE_BENCHMARK,
          "benchmark_file": input("  Enter benchmark file path [default: " + OdatixSettings.DEFAULT_BENCHMARK_FILE + "]: ") or OdatixSettings.DEFAULT_BENCHMARK_FILE,
          "clean_settings_file": input("  Enter clean settings file [default: " + OdatixSettings.DEFAULT_CLEAN_SETTINGS_FILE + "]: ") or OdatixSettings.DEFAULT_CLEAN_SETTINGS_FILE,
          "simulation_settings_file": input("  Enter simulation settings file [default: " + OdatixSettings.DEFAULT_SIMULATION_SETTINGS_FILE + "]: ") or OdatixSettings.DEFAULT_SIMULATION_SETTINGS_FILE,
          "fmax_synthesis_settings_file": input("  Enter fmax synthesis settings file [default: " + OdatixSettings.DEFAULT_FMAX_SYNTHESIS_SETTINGS_FILE + "]: ") or OdatixSettings.DEFAULT_FMAX_SYNTHESIS_SETTINGS_FILE,
          # "range_synthesis_settings_file": input("  Enter range synthesis settings file [default: " + OdatixSettings.DEFAULT_RANGE_SYNTHESIS_SETTINGS_FILE + "]: ") or OdatixSettings.DEFAULT_RANGE_SYNTHESIS_SETTINGS_FILE
        }

        with open(settings_filename, "w") as f:
          yaml.dump(settings_data, f, sort_keys=False)

        printc.say("Odatix settings file \"" + settings_filename + "\" has been generated.", script_name=script_name)
        print()
        return True

  @staticmethod
  def init_examples():
    try:
      src_path = OdatixSettings.odatix_examples_path
      dst_path = os.getcwd()
      copytree(src_path, dst_path, dirs_exist_ok=True)
    except Exception as e:
      printc.error("Could not copy examples: " + str(e), script_name=script_name)
      return False
    return True

  @staticmethod
  def init_path():
    try:
      src_path = OdatixSettings.odatix_init_path
      dst_path = os.getcwd()
      copytree(src_path, dst_path, dirs_exist_ok=True)
    except Exception as e:
      printc.error("Could not initialize directory: " + str(e), script_name=script_name)
      return False
    return True

  @staticmethod
  def init_directory_dialog(include_examples=None, prog=""):
    printc.note("This command will create all the configuration files needed by Odatix in the current dictory.", script_name=script_name)
    printc.warning("This will overwrite any existing configuration files.", script_name=script_name)
    printc.say("Would you like to continue? ", end="", script_name=script_name)
    answer = ask_yes_no()
    if answer == False:
      return False
    success = OdatixSettings.init_path()
    if not success:
      return False
    if include_examples is None:
      printc.say("Would you like to add Odatix examples? ", end="", script_name=script_name)
      include_examples = ask_yes_no()
    if include_examples:
      success = OdatixSettings.init_examples()
      if not success:
        return False
    OdatixSettings.init_success(prog)
    return True

  @staticmethod
  def init_directory_nodialog(include_examples=None, prog=""):
    success = OdatixSettings.init_path()
    if not success:
      return False
    if include_examples:
      success = OdatixSettings.init_examples()
      if not success:
        return False
    OdatixSettings.init_success(prog)
    return True

  @staticmethod
  def init_success(prog=""):
    printc.green("Your directory can now be used by Odatix!", script_name=script_name)
    printc.say("Run ", end="", script_name=script_name)
    printc.bold(prog + " -h", end="")
    print(" to get a list of useful commands")



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re

def edit_config_file(arch, config_file): 
  with open(config_file, 'r') as f:
    cf_content = f.read()
    cf_content = re.sub("(set tmp_path.*)",           "set tmp_path           " + os.path.realpath(arch.tmp_dir), cf_content)
    cf_content = re.sub("(set script_path.*)",        "set script_path        " + os.path.realpath(arch.tmp_script_path), cf_content)
    cf_content = re.sub("(set report_path.*)",        "set report_path        " + os.path.realpath(arch.tmp_report_path), cf_content)
    cf_content = re.sub("(set rtl_path.*)",           "set rtl_path           " + arch.rtl_path, cf_content)
    cf_content = re.sub("(set arch_path.*)",          "set arch_path          " + arch.arch_path, cf_content)
    cf_content = re.sub("(set clock_signal.*)",       "set clock_signal       " + arch.clock_signal, cf_content)
    cf_content = re.sub("(set reset_signal.*)",       "set reset_signal       " + arch.reset_signal, cf_content)
    cf_content = re.sub("(set top_level_module.*)",   "set top_level_module   " + arch.top_level_module, cf_content)
    cf_content = re.sub("(set top_level_file.*)",     "set top_level_file     " + arch.top_level_filename, cf_content)
    cf_content = re.sub("(set fmax_lower_bound.*)",   "set fmax_lower_bound   " + arch.fmax_lower_bound, cf_content)
    cf_content = re.sub("(set fmax_upper_bound.*)",   "set fmax_upper_bound   " + arch.fmax_upper_bound, cf_content)
    cf_content = re.sub("(set lib_name.*)",           "set lib_name           " + arch.lib_name, cf_content)
    cf_content = re.sub("(set constraints_file.*)",   "set constraints_file   $tmp_path/" + arch.constraint_filename, cf_content)
 
  with open(config_file, 'w') as f:
    f.write(cf_content)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import curses
import re


A_NORMAL = 0


class AnsiToCursesConverter:
  LIGHT_OFFSET = 8

  # Define ANSI codes for colors and attributes
  ANSI_COLORS = {
    "30": curses.COLOR_BLACK,
    "31": curses.COLOR_RED,
    "32": curses.COLOR_GREEN,
    "33": curses.COLOR_YELLOW,
    "34": curses.COLOR_BLUE,
    "35": curses.COLOR_MAGENTA,
    "36": curses.COLOR_CYAN,
    "37": curses.COLOR_WHITE,
    "90": curses.COLOR_BLACK + LIGHT_OFFSET,  # Light black (grey)
    "91": curses.COLOR_RED + LIGHT_OFFSET,  # Light red
    "92": curses.COLOR_GREEN + LIGHT_OFFSET,  # Light green
    "93": curses.COLOR_YELLOW + LIGHT_OFFSET,  # Light yellow
    "94": curses.COLOR_BLUE + LIGHT_OFFSET,  # Light blue
    "95": curses.COLOR_MAGENTA + LIGHT_OFFSET,  # Light magenta
    "96": curses.COLOR_CYAN + LIGHT_OFFSET,  # Light cyan
    "97": curses.COLOR_WHITE + LIGHT_OFFSET,  # Light white
  }

  def __init__(self):
    self.current_color = None
    self.current_intensity = 0
    self.initialized = False

  def initialize_colors(self):
    if not self.initialized:
      curses.start_color()
      curses.use_default_colors()
      for code, color in AnsiToCursesConverter.ANSI_COLORS.items():
        if color != -1:
          curses.init_pair(int(code), color, -1)
      self.current_color = curses.color_pair(0)
      self.initialized = True

  def add_ansi_str(self, win, text, debug_win=None, width=-1):
    self.initialize_colors()

    # Regular expression to find ANSI escape sequences
    ansi_escape = re.compile(r"\x1b\[([0-9;]*)m")

    # Find all ANSI codes and normal text
    segments = ansi_escape.split(text)

    current_width = 0
    truncated = False

    for i, segment in enumerate(segments):
      if i % 2 == 0:
        # Normal text segment
        if width > 0:
          remaining_width = width - current_width
          if remaining_width > 0:
            if len(segment) > remaining_width:
              segment = segment[:remaining_width-3]
              truncated = True
            current_width += len(segment)
            win.addstr(segment, self.current_color | self.current_intensity)
          if truncated:
            win.addstr("...", curses.color_pair(90))
        else:
          win.addstr(segment, self.current_color | self.current_intensity)
      else:
        # ANSI code segment
        codes = segment.split(";")
        for code in codes:
          if code in AnsiToCursesConverter.ANSI_COLORS:
            color_pair = curses.color_pair(int(code))
            self.current_color = color_pair  # Use the new color
          elif code == "0":  # Reset code
            self.current_color = curses.color_pair(0)  # Reset to default color and attributes
            self.current_intensity = A_NORMAL
          elif code == "1":  # Bold code
            self.current_intensity = curses.A_BOLD
          elif code == "22":  # Normal intensity
            self.current_intensity = A_NORMAL
          else:
            if debug_win is not None:
              debug_win.addstr("{Unsupported ANSI escape code:" + code + "}", curses.color_pair(31))

    win.refresh()
    if debug_win is not None:
      debug_win.refresh()

  def test(self, stdscr):
    stdscr.clear()
    text = " default_text\n \x1b[1m\x1b[31mbold_red_text\x1b[0m\n \x1b[34mblue_and_\x1b[32mgreen_text\n \x1b[31mred_text\n \x1b[0mdefault_text\n \x1b[52merror_text"
    height, width = stdscr.getmaxyx()

    # Ensure the text fits within the window width
    self.add_ansi_str(stdscr, text, width=width)

    # Wait for a key press to exit
    stdscr.getch()


if __name__ == "__main__":
  conv = AnsiToCursesConverter()
  curses.wrapper(conv.test)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import yaml
import glob
import shutil
import argparse

import odatix.lib.printc as printc
from odatix.lib.utils import read_from_list, KeyNotInListError, BadValueInListError
from odatix.lib.settings import OdatixSettings

script_name = os.path.basename(__file__)

DANGEROUS_PATHS = ["/", "./", "./*", "*", "~", ".", ".."]

######################################
# Parse Arguments
######################################

def add_arguments(parser):
  parser.add_argument("-i", "--input", default=None, help="input settings file")
  parser.add_argument("-f", "--force", action="store_true", help="force delete (dangerous!)")
  parser.add_argument("-v", "--verbose", action="store_true", help="print extra details")
  parser.add_argument("-q", "--quiet", action="store_true", help="do not print anything, except errors")
  parser.add_argument(
    "-c",
    "--config",
    default=OdatixSettings.DEFAULT_SETTINGS_FILE,
    help="global settings file for Odatix (default: " + OdatixSettings.DEFAULT_SETTINGS_FILE + ")",
  )

def parse_arguments():
  parser = argparse.ArgumentParser(description="Clean up current directory")
  add_arguments(parser)
  return parser.parse_args()

######################################
# Helper Functions
######################################

def remove_path(path, force=False, verbose=False, quiet=False):
      
  cwd = os.getcwd()
  full_path = os.path.realpath(os.path.join(cwd, path))

  if path in DANGEROUS_PATHS or full_path in DANGEROUS_PATHS:
    if not force:
      if not quiet:
        printc.warning("Deleting \"" + full_path + "\" seams dangerous! Use --force to force deletion (use at your own risks!)", script_name=script_name)
      return

  # Handle asterisks
  paths_to_remove = glob.glob(full_path)

  # Check if the lis is empty
  if not paths_to_remove:
    if verbose:
      printc.warning("Path \"" + full_path + "\" does not exist or is not accessible.", script_name=script_name)
    return

  # Remove path
  for p in paths_to_remove:
    try:
      if os.path.isfile(p) or os.path.islink(p):
        os.remove(p)
        if not quiet:
          printc.say("Removed \"" + p + "\" (file)", script_name=script_name)
      elif os.path.isdir(p):
        shutil.rmtree(p)
        if not quiet:
          printc.say("Removed \"" + p + "\" (directory)", script_name=script_name)
      else:
        if verbose:
          printc.warning("Path \"" + p + "\" does not exist or is not accessible.", script_name=script_name)
    except Exception as e:
      printc.error("Failed to remove \"" + p + "\".", script_name=script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))

######################################
# Clean
######################################

def clean(settings_filename, force=False, verbose=False, quiet=False):
  if not os.path.isfile(settings_filename):
    if not quiet:
      printc.note("There is no clean settings file \"" + settings_filename + "\" in \"" + os.path.realpath(".") + "\". Using default Odatix clean settings file", script_name)
    settings_filename = os.path.join(OdatixSettings.odatix_path, OdatixSettings.DEFAULT_CLEAN_SETTINGS_FILE)
    if not os.path.isfile(settings_filename):
      printc.error("There is no default Odatix clean settings file \"" + settings_filename, script_name)
      sys.exit(-1)
  with open(settings_filename, "r") as f:
    try:
      settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    except Exception as e:
      printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
      printc.cyan("error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)
    try:
      remove_list = read_from_list("remove_list", settings_data, settings_filename, script_name=script_name)
    except (KeyNotInListError, BadValueInListError):
      sys.exit(-1)

    if not isinstance(remove_list, list):
      printc.error("\"remove_list\" from settings file \"" + settings_filename + "\" is not a list", script_name)
      printc.note("Are you missing dashes (-)?", script_name)
      sys.exit(-1)

    for path in remove_list:
      remove_path(path, force, verbose, quiet)

######################################
# Main
######################################

def main(args, settings=None): 
  # Get settings
  if settings is None:
    settings = OdatixSettings(args.config)
    if not settings.valid:
      sys.exit(-1)

  if args.input is None:
    args.input = settings.clean_settings_file

  clean(
    settings_filename=args.input,
    force=args.force,
    verbose=args.verbose,
    quiet=args.quiet
  )

if __name__ == "__main__":
  args = parse_arguments()
  main(args)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import shutil
import platform
import traceback
from datetime import datetime

import odatix.lib.printc as printc

YAML_BOOL = ('true', 'false', 'yes', 'no', 'on', 'off')

# python 3.8+ like copytree
def copytree(src, dst, dirs_exist_ok=False, **kwargs):
  if not os.path.exists(dst):
    shutil.copytree(src, dst, **kwargs)
  else:
    if dirs_exist_ok:
      for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
          shutil.rmtree(d, ignore_errors=True)
          shutil.copytree(s, d, **kwargs)
        else:
          shutil.copy2(s, d)
    else:
      raise

def chunk_list(lst, n):
  for i in range(0, len(lst), n):
    yield lst[i:i + n]

class KeyNotInListError(Exception):
  pass

class BadValueInListError(Exception):
  pass

def read_from_list(key, input_list, filename, raise_if_missing=True, optional=False, print_error=True, parent=None, type=None, script_name=""):
  parent_string = "" if parent == None else ", inside list \"" + parent + "\","
  if key in input_list:
    value = input_list[key]
    if type is None:
      return value
    else:
      if isinstance(value, type):
        return value
      else:
        if print_error:
          if optional:
            printc.note("Value \"" + str(value) + "\" for key \"" + key + "\"" + parent_string + " in \"" + filename + "\" is of type \"" + value.__class__.__name__ + "\" while it should be of type \"" + type.__name__ + "\". Using default values instead.", script_name)
          else:
            printc.error("Value \"" + str(value) + "\" for key \"" + key + "\"" + parent_string + " in \"" + filename + "\" is of type \"" + value.__class__.__name__ + "\" while it should be of type \"" + type.__name__ + "\".", script_name)
        if raise_if_missing:
          raise BadValueInListError
        return False
  else:
    if print_error:
      if optional:
        printc.note("Cannot find optional key \"" + key + "\"" + parent_string + " in \"" + filename + "\". Using default values instead.", script_name)
      else:
        printc.error("Cannot find key \"" + key + "\"" + parent_string + " in \"" + filename + "\".", script_name)
    if raise_if_missing:
      raise KeyNotInListError
    return False

def read_from_config(identifier, config, filename, script_name=""):
  if identifier in config[settings_ini_section]:
    return config[settings_ini_section][identifier]
  else:
    printc.error("Cannot find identifier \"" + identifier + "\" in \"" + filename + "\".", script_name)
    raise
    return False

def move_cursor_up():
  sys.stdout.write('\x1b[1A') # Move cursor up
  sys.stdout.write("\033[K") # Clear to the end of line
  sys.stdout.flush()

def progress_bar(progress, title, title_size=50, bar_size=50, endstr='', progress_char="#"):
  if progress is None:
    limit = 1
  else:
    if progress > 100:
      progress = 100
    limit = int(progress * bar_size / 100)

  padded_title = title.ljust(title_size)

  print(padded_title + " [", end='')
  if progress is None:
    printc.color(printc.colors.BLINK)
  for i in range(0, limit):
    print(progress_char, end='')
  for i in range(limit, bar_size):
    print(' ', end='')
  if progress is None:
    printc.color(printc.colors.ENDC)
  print("]", end='')
  if progress is not None:
    print(" {}%".format(int(progress)) + endstr)
  else:
    print()

def ask_to_continue(exit_code=-1):
  print("Continue? ", end="")
  answer = ask_yes_no()
  if answer is False:
    sys.exit(exit_code)
    
def ask_yes_no():
  while True:
    answer = input("(Y/n) ")
    if answer.lower() in ['yes', 'ye', 'y', '1', '']:
      return True
    elif answer.lower() in ['no', 'n', '0']:
      return False
    else:
      print("Please enter yes or no")

def create_dir(dir):
  if os.path.isdir(dir):
    shutil.rmtree(dir)
  os.makedirs(dir)

def internal_error(e, error_logfile, script_name):
  tb_full = traceback.format_exc()
  command_line = ' '.join(sys.argv)
  current_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
  system_info = {
    "OS": platform.system(),
    "OS Version": platform.version(),
    "Python Version": platform.python_version(),
    "Machine": platform.machine(),
  }
  with open(error_logfile, "w") as log_file:
    log_file.write("Date and Time: " + current_time + "\n\n")
    log_file.write("System Information:\n")
    for key, value in system_info.items():
      log_file.write("  " + key + ": " + value + "\n")
    log_file.write("\nCommand:\n")
    log_file.write("  " + command_line + "\n\n")
    log_file.write(tb_full)
  printc.internal_error(type(e).__name__ + ": " + str(e), script_name)
  printc.note('Full error details written to "' + error_logfile + '"', script_name)
  printc.note("Please, report this error with the error log attached", script_name)

def safe_df_append(df, row, ignore_index=True):
  if hasattr(df, 'append'):
    return df.append(row, ignore_index=ignore_index)
  else:
    return df._append(row, ignore_index=ignore_index)


# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import math
import yaml

from os.path import isfile
from os.path import isdir

from odatix.lib.architecture_handler import ArchitectureHandler
from odatix.lib.utils import *

script_name = os.path.basename(__file__)

class Simulation:
  def __init__(self, sim_name, sim_display_name, architecture, tmp_dir, source_sim_dir, override_parameters, override_param_target_filename, override_param_filename, override_start_delimiter, override_stop_delimiter, simulation_command):
    self.sim_name = sim_name
    self.sim_display_name = sim_display_name
    self.architecture = architecture
    self.tmp_dir = tmp_dir
    self.source_sim_dir = source_sim_dir
    self.override_parameters = override_parameters
    self.override_param_target_filename = override_param_target_filename
    self.override_param_filename = override_param_filename
    self.override_start_delimiter = override_start_delimiter
    self.override_stop_delimiter = override_stop_delimiter
    self.simulation_command = simulation_command
    

class SimulationHandler:

  def __init__(self, work_path, arch_path, sim_path, work_script_path, log_path, param_settings_filename, sim_settings_filename, sim_makefile_filename, overwrite):
    self.work_path = work_path
    self.arch_path = arch_path
    self.sim_path = sim_path
    self.work_script_path = work_script_path
    self.log_path = log_path
    self.overwrite = overwrite
    self.param_settings_filename = param_settings_filename
    self.sim_settings_filename = sim_settings_filename
    self.sim_makefile_filename = sim_makefile_filename
    self.reset_lists()

  def reset_lists(self):
    self.no_settings_sims = []
    self.banned_sim_param = []
    self.valid_sims = []
    self.cached_sims = []
    self.overwrite_sims = []
    self.error_sims = []
    self.incomplete_sims = []
    self.new_sims = []

  def get_simulations(self, simulations):

    self.reset_lists()
    self.simulation_instances = []

    arch_handler = ArchitectureHandler(
      work_path = self.work_path,
      arch_path = self.arch_path,
      script_path = "",
      work_script_path = self.work_script_path,
      work_report_path = "",
      log_path = self.log_path,
      process_group=True,
      eda_target_filename = "",
      fmax_status_filename = "",
      frequency_search_filename = "",
      param_settings_filename = self.param_settings_filename,
      valid_status = "",
      valid_frequency_search = "",
      default_fmax_lower_bound = 0,
      default_fmax_upper_bound = 1000,
      overwrite = self.overwrite
    )

    if simulations is not None:
      for sim_dict in simulations:
        if sim_dict is not None:
          for sim, arch_list in sim_dict.items():
            if arch_list is not None and arch_list is not None:
              for arch in arch_list:
                simulation_instance = self.get_simulation(sim, arch, arch_handler)
                if simulation_instance is not None:
                  self.simulation_instances.append(simulation_instance)

    return self.simulation_instances
    
  
  def get_simulation(self, sim, arch, arch_handler):
    tmp_dir = self.work_path + '/' + sim + '/' + arch 

    sim_name = sim
    sim_display_name = sim + " (" + arch + ")"
    simulation_command = None

    # check if sim has been banned
    if sim in self.banned_sim_param:
      self.error_sims.append(sim_display_name)
      return None

    # check if sim dir exists
    source_sim_dir = self.sim_path + '/' + sim
    if not isdir(source_sim_dir):
      printc.error("There is no directory \"" + sim + "\" in directory \"" + self.sim_path + "\"", script_name)
      self.banned_sim_param.append(sim)
      self.error_sims.append(sim_display_name)
      return None

    # check if sim dir exists
    source_sim_dir = self.sim_path + '/' + sim
    if not isdir(source_sim_dir):
      printc.error("There is no directory \"" + sim + "\" in directory \"" + self.sim_path + "\"", script_name)
      self.banned_sim_param.append(sim)
      self.error_sims.append(sim_display_name)
      return None

    # get architecture
    architecture = arch_handler.get_architecture(arch)
    if architecture is None:
      self.error_sims.append(sim_display_name)
      return None

    # check if makefile exists
    makefile_filename = source_sim_dir + '/' + self.sim_makefile_filename
    if not isfile(makefile_filename):
      printc.error("There is no setting \"Makefile\" in directory \"" + source_sim_dir + "\"", script_name)
      printc.note("A Makefile with a rule \"sim\" is mandatory", script_name)
      self.banned_sim_param.append(sim)
      self.error_sims.append(sim_display_name)
      return None

    override_parameters = False
    override_param_target_filename = ""
    override_param_file = ""
    override_start_delimiter = ""
    override_stop_delimiter = ""

    # check if settings file exists
    if sim not in self.no_settings_sims:
      settings_filename = source_sim_dir + '/' + self.sim_settings_filename
      if not isfile(settings_filename):
        printc.note("There is no setting file \"" + self.sim_settings_filename + "\" in directory \"" + source_sim_dir + "\"", script_name)
        self.no_settings_sims.append(sim)
      else:
        with open(settings_filename, 'r') as f:
          try:
            settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
          except Exception as e:
            printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
            printc.cyan("error details: ", end="", script_name=script_name)
            print(str(e))
            self.banned_arch_param.append(arch_param_dir)
            self.error_archs.append(arch_display_name)
            return None # if an identifier is missing

          # get use_parameters, start_delimiter and stop_delimiter
          use_parameters, start_delimiter, stop_delimiter = arch_handler.get_use_parameters(arch, arch, settings_data, settings_filename, add_to_error_list=False)
          if use_parameters is None:
            self.banned_sim_param.append(sim)
            self.error_sims.append(sim_display_name)
            return None
          elif start_delimiter is None or stop_delimiter is None:
            self.error_sims.append(sim_display_name)
            return None

          # overwrite architecture settings
          architecture.use_parameters = use_parameters
          architecture.start_delimiter = start_delimiter
          architecture.stop_delimiter = stop_delimiter

          # get param_target_file
          if use_parameters:
            try:
              param_target_filename = read_from_list('param_target_file', settings_data, settings_filename, script_name=script_name)
              # check if param target file path exists
              param_target_file_rtl = architecture.rtl_path + '/' + param_target_filename
              param_target_file_sim = source_sim_dir + '/' + param_target_filename
              #if not isfile(param_target_file_rtl) and not isfile(param_target_file_sim): 
                #printc.warning("The parameter target file \"" + param_target_filename + "\" specified in \"" + settings_filename + "\" does not seem to exist", script_name)
              # overwrite architecture settings
              architecture.param_target_filename = param_target_filename
            except (KeyNotInListError, BadValueInListError):
              self.banned_sim_param.append(sim)
              self.error_sims.append(sim_display_name)
              return None

          # get override_parameters
          try:
            override_parameters = read_from_list('override_parameters', settings_data, settings_filename, type=bool, script_name=script_name)
          except (KeyNotInListError, BadValueInListError):
            self.banned_sim_param.append(sim)
            self.error_sims.append(sim_display_name)
            return None

          if override_parameters:
            # get override_param_target_file
            try:
              override_param_file = read_from_list('override_param_file', settings_data, settings_filename, print_error=False, script_name=script_name)
            except (KeyNotInListError, BadValueInListError):
              printc.error("Cannot find key \"override_param_file\" in \"" + settings_filename + "\", while \"override_parameters\" is true", script_name)
              self.banned_sim_param.append(sim)
              self.error_sims.append(sim_display_name)
              return None

            # check if parameter file exists
            if not isfile(source_sim_dir + '/' + override_param_file):
              printc.error("There is no parameter file \"" + source_sim_dir + '/' + override_param_file + "\", while override_parameters=true", script_name)
              self.error_sims.append(sim_display_name)
              return True
          
          if override_parameters:
            # get start delimiter
            try:
              override_start_delimiter = read_from_list('override_start_delimiter', settings_data, settings_filename, print_error=False, script_name=script_name)
            except (KeyNotInListError, BadValueInListError):
              printc.error("Cannot find key \"override_start_delimiter\" in \"" + settings_filename + "\", while \"override_parameters\" is true", script_name)
              self.banned_sim_param.append(sim)
              self.error_sims.append(sim_display_name)
              return None

            # get stop delimiter
            try:
              override_stop_delimiter = read_from_list('override_stop_delimiter', settings_data, settings_filename, print_error=False, script_name=script_name)
            except (KeyNotInListError, BadValueInListError):
              printc.error("Cannot find key \"override_stop_delimiter\" in \"" + settings_filename + "\", while \"override_parameters\" is true", script_name)
              self.banned_sim_param.append(sim)
              self.error_sims.append(sim_display_name)
          else:
            override_start_delimiter = ""
            override_stop_delimiter = ""

          # get override_param_target_file
          if override_parameters:
            try:
              override_param_target_filename = read_from_list('override_param_target_file', settings_data, settings_filename, script_name=script_name)

              # check if param target file path exists
              override_param_target_file_rtl = architecture.rtl_path + '/' + override_param_target_filename
              override_param_target_file_sim = source_sim_dir + '/' + override_param_target_filename
              #if not isfile(override_param_target_file_rtl) and not isfile(override_param_target_file_sim): 
                #printc.warning("The override parameter target file \"" + override_param_target_filename + "\" specified in \"" + settings_filename + "\" does not seem to exist", script_name)
            except (KeyNotInListError, BadValueInListError):
              self.banned_sim_param.append(sim)
              self.error_sims.append(sim_display_name)
              return None
          else:
            override_param_target_filename = "/dev/null"
      
    # check if the architecture is in cache and has a status file
    if isdir(tmp_dir):
      if self.overwrite:
        printc.warning("Found cached results for \"" + sim_display_name +"\".", script_name)
        self.overwrite_sims.append(sim_display_name)
      else:
        printc.note("Found cached results for \"" + sim_display_name + "\". Skipping.", script_name)
        self.cached_sims.append(sim_display_name)
        return None
    else:
      self.new_sims.append(sim_display_name)

    # passed all check: added to the list
    self.valid_sims.append(sim_display_name)

    sim_instance = Simulation(
      sim_name = sim_name,
      sim_display_name = sim_display_name,
      architecture = architecture,
      tmp_dir = tmp_dir,
      source_sim_dir = source_sim_dir,
      override_parameters = override_parameters,
      override_param_target_filename = override_param_target_filename,
      override_param_filename = override_param_file,
      override_start_delimiter = override_start_delimiter,
      override_stop_delimiter = override_stop_delimiter,
      simulation_command = simulation_command
    )

    return sim_instance

  def print_summary(self):
    SimulationHandler.print_sim_list(self.new_sims, "New simulations", printc.colors.ENDC)
    SimulationHandler.print_sim_list(self.incomplete_sims, "Incomplete results (will be overwritten)", printc.colors.YELLOW)
    SimulationHandler.print_sim_list(self.cached_sims, "Existing results (skipped)", printc.colors.CYAN)
    SimulationHandler.print_sim_list(self.overwrite_sims, "Existing results (will be overwritten)", printc.colors.YELLOW)
    SimulationHandler.print_sim_list(self.error_sims, "Invalid settings, (skipped, see errors above)", printc.colors.RED)

  def get_chuncks(self, nb_jobs):
    if len(self.simulation_instances) > nb_jobs:
      nb_chunks = math.ceil(len(self.simulation_instances) / nb_jobs)
      print()
      printc.note("Current maximum number of jobs is " + str(nb_jobs) + ". Simulations will be split in " + str(nb_chunks) + " chunks")
      self.simulation_instances_chunks = list(chunk_list(self.simulation_instances, nb_jobs))
    else:
      nb_chunks = 1
      self.simulation_instances_chunks = []
    return self.simulation_instances_chunks, nb_chunks


  def get_valid_sim_count(self):
    return len(self.valid_sims)

  @staticmethod
  def print_sim_list(arch_list, description, color):
    if not len(arch_list) > 0:
      return

    print()
    printc.bold(description + ":")
    for arch in arch_list:
      printc.color(color)
      print("  - " + arch)
    printc.endc()
